# Copyright OTT-JAX
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import functools
from typing import Any, Callable, Optional, Tuple, Union

import jax
import jax.numpy as jnp
import jax.tree_util as jtu
import numpy as np

from ott import utils
from ott.geometry import costs, pointcloud
from ott.problems.linear import linear_problem
from ott.solvers import linear
from ott.solvers.linear import sinkhorn

__all__ = [
    "sort", "ranks", "sort_with", "quantile", "quantile_normalization",
    "quantize", "topk_mask", "multivariate_cdf_quantile_maps"
]

Func_t = Callable[[jnp.ndarray], jnp.ndarray]


def transport_for_sort(
    inputs: jnp.ndarray,
    weights: Optional[jnp.ndarray] = None,
    target_weights: Optional[jnp.ndarray] = None,
    squashing_fun: Optional[Callable[[jnp.ndarray], jnp.ndarray]] = None,
    epsilon: float = 1e-2,
    **kwargs: Any,
) -> sinkhorn.SinkhornOutput:
  r"""Solve reg. OT, from inputs to a weighted family of increasing values.

  Args:
    inputs: Array[num_points]. Must be one dimensional.
    weights: Array[num_points]. Weight vector `a` for input values.
    target_weights: Array[num_targets]: Weight vector of the target
      measure. It may be of different size than `weights`.
    squashing_fun: function taking an array to squash all its entries in [0,1].
      sigmoid of whitened values by default. Can be set to be the identity by
      passing ``squashing_fun = lambda x : x`` instead.
    epsilon: the regularization parameter.
    kwargs: keyword arguments for
      :class:`~ott.solvers.linear.sinkhorn.Sinkhorn`.

  Returns:
    A :class:`~ott.solvers.linear.sinkhorn.SinkhornOutput` object.
  """
  shape = inputs.shape
  if len(shape) > 2 or (len(shape) == 2 and shape[1] != 1):
    raise ValueError(
        f"Shape ({shape}) not supported. The input should be one-dimensional."
    )

  x = jnp.expand_dims(jnp.squeeze(inputs), axis=1)
  if squashing_fun is None:
    squashing_fun = lambda z: jax.nn.sigmoid((z - jnp.mean(z)) /
                                             (jnp.std(z) + 1e-10))
  x = squashing_fun(x)

  a = jnp.squeeze(weights)
  b = jnp.squeeze(target_weights)
  num_targets = inputs.shape[0] if b is None else b.shape[0]
  y = jnp.linspace(0.0, 1.0, num_targets)[:, jnp.newaxis]

  geom = pointcloud.PointCloud(x, y, epsilon=epsilon)
  prob = linear_problem.LinearProblem(geom, a=a, b=b)

  solver = sinkhorn.Sinkhorn(**kwargs)

  return solver(prob)


def apply_on_axis(op, inputs, axis, *args, **kwargs: Any) -> jnp.ndarray:
  """Apply a differentiable operator on a given axis of the input.

  Args:
    op: a differentiable operator (can be ranks, quantile, etc.)
    inputs: Array of any shape.
    axis: the axis (int) or tuple of ints on which to apply the operator. If
      several axes are passed to the operator, those are merged as a single
      dimension.
    args: other positional arguments to the operator.
    kwargs: other positional arguments to the operator.

  Returns:
    An Array holding the output of the differentiable operator on the given
    axis.
  """
  op_inner = functools.partial(op, **kwargs)
  axis = (axis,) if isinstance(axis, int) else axis
  num_points = np.prod(np.array(inputs.shape)[(axis,)])
  permutation = np.arange(len(inputs.shape))
  axis = tuple(permutation[a] for a in axis)
  permutation = tuple(sorted(set(permutation) - set(axis)) + sorted(axis))
  inputs = jnp.transpose(inputs, permutation)

  batch_fn = jax.vmap(op_inner, in_axes=(0,) + (None,) * len(args))
  result = batch_fn(jnp.reshape(inputs, (-1, num_points)), *args)
  shrink = len(axis)
  result = jnp.reshape(result, inputs.shape[:-shrink] + result.shape[-1:])

  permutation = tuple(range(len(result.shape)))
  rank = len(result.shape) - 1
  axis = min(axis)
  permutation = permutation[:axis] + (rank,) + permutation[axis:-1]
  return jnp.transpose(result, permutation)


def _sort(
    inputs: jnp.ndarray, topk: int, num_targets: Optional[int], **kwargs: Any
) -> jnp.ndarray:
  """Apply the soft sort operator on a one dimensional array."""
  num_points = inputs.shape[0]
  a = jnp.ones((num_points,)) / num_points
  if 0 < topk < num_points:
    start_index = 1
    b = jnp.concatenate([
        jnp.array([(num_points - topk) / num_points]),
        jnp.ones(topk) / num_points
    ])
  else:
    # Use the "sorting" initializer if default uniform weights of same size.
    if num_targets is None or num_targets == num_points:
      num_targets = num_points
      # use sorting initializer in this case.
      kwargs.setdefault("initializer", "sorting")
    start_index = 0
    b = jnp.ones((num_targets,)) / num_targets
  ot = transport_for_sort(inputs, a, b, **kwargs)
  out = 1.0 / b * ot.apply(inputs, axis=0)
  return out[start_index:]


def sort(
    inputs: jnp.ndarray,
    axis: int = -1,
    topk: int = -1,
    num_targets: Optional[int] = None,
    **kwargs: Any,
) -> jnp.ndarray:
  r"""Apply the soft sort operator on a given axis of the input.

  For instance:

  .. code-block:: python

    x = jax.random.uniform(rng, (100,))
    x_sorted = sort(x)

  will output sorted convex-combinations of values contained in ``x``, that are
  differentiable approximations to the sorted vector of entries in ``x``.
  These can be compared with the values produced by :func:`jax.numpy.sort`,

  .. code-block:: python

    x_sorted = jax.numpy.sort(x)

  Args:
    inputs: Array of any shape.
    axis: the axis on which to apply the soft-sorting operator.
    topk: if set to a positive value, the returned vector will only contain
      the top-k values. This also reduces the complexity of soft-sorting, since
      the number of target points to which the slice of the ``inputs`` tensor
      will be mapped to will be equal to ``topk + 1``.
    num_targets: if ``topk`` is not specified, a vector of size``num_targets``
      is returned. This defines the number of (composite) sorted values computed
      from the inputs (each value is a convex combination of values recorded in
      the inputs, provided in increasing order). If neither ``topk`` nor
      ``num_targets`` are specified, ``num_targets`` defaults to the size of the
      slices of the input that are sorted, i.e. ``inputs.shape[axis]``, and the
      number of composite sorted values is equal to the slice of the inputs that
      are sorted. As a result, the output is of the same size as ``inputs``.
    kwargs: keyword arguments passed on to lower level functions. Of interest
      to the user are ``squashing_fun``, which will redistribute the values in
      ``inputs`` to lie in :math:`[0,1]` (sigmoid of whitened values by default)
      to solve the optimal transport problem;
      :class:`cost_fn <ott.geometry.costs.CostFn>` object of
      :class:`~ott.geometry.pointcloud.PointCloud`, which defines the ground
      1D cost function to transport from ``inputs`` to the ``num_targets``
      target values; ``epsilon`` regularization parameter. Remaining ``kwargs``
      are passed on to parameterize the
      :class:`~ott.solvers.linear.sinkhorn.Sinkhorn` solver.

  Returns:
    An Array of the same shape as the input, except on ``axis``, where that size
    will be equal to ``topk`` or ``num_targets``, with soft-sorted values on the
    given axis. Same size as ``inputs`` if both these parameters are ``None``.
  """
  return apply_on_axis(_sort, inputs, axis, topk, num_targets, **kwargs)


def _ranks(
    inputs: jnp.ndarray, num_targets, target_weights, **kwargs: Any
) -> jnp.ndarray:
  """Apply the soft ranks operator on a one dimensional array."""
  num_points = inputs.shape[0]
  if target_weights is None:
    num_targets = num_points if num_targets is None else num_targets
    target_weights = jnp.ones((num_targets,)) / num_targets
  else:
    num_targets = target_weights.shape[0]
  a = jnp.ones((num_points,)) / num_points
  ot = transport_for_sort(inputs, a, target_weights, **kwargs)
  out = 1.0 / a * ot.apply(jnp.arange(num_targets, dtype=inputs.dtype), axis=1)
  out *= (num_points - 1.0) / (num_targets - 1.0)
  return jnp.reshape(out, inputs.shape)


def ranks(
    inputs: jnp.ndarray,
    axis: int = -1,
    num_targets: Optional[int] = None,
    target_weights: Optional[jnp.ndarray] = None,
    **kwargs: Any,
) -> jnp.ndarray:
  r"""Apply the soft rank operator on input tensor.

  For instance:

  .. code-block:: python

    x = jax.random.uniform(rng, (100,))
    x_ranks = ranks(x)

  will output values that are differentiable approximations to the ranks of
  entries in ``x``. These should be compared to the non-differentiable rank
  vectors, namely the normalized inverse permutation produced by
  :func:`jax.numpy.argsort`, which can be obtained as:

  .. code-block:: python

    x_ranks = jax.numpy.argsort(jax.numpy.argsort(x))

  Args:
    inputs: Array of any shape.
    axis: the axis on which to apply the soft-sorting operator.
    target_weights: This vector contains weights (summing to 1) that describe
      amount of mass shipped to targets.
    num_targets: If ``target_weights`  is ``None``, ``num_targets`` is
      considered to define the number of targets used to rank inputs. Each
      rank in the output will be a convex combination of
      ``{1, .., num_targets}``. The weight of each of these points
      is assumed to be uniform. If neither ``num_targets`` nor
      ``target_weights`` are specified, ``num_targets`` defaults to the size
      of the slices of the input that are sorted, i.e. ``inputs.shape[axis]``.
    kwargs: keyword arguments passed on to lower level functions. Of interest
      to the user are ``squashing_fun``, which will redistribute the values in
      ``inputs`` to lie in :math:`[0,1]` (sigmoid of whitened values by default)
      to solve the optimal transport problem;
      :class:`cost_fn <ott.geometry.costs.CostFn>` object of
      :class:`~ott.geometry.pointcloud.PointCloud`, which defines the ground
      1D cost function to transport from ``inputs`` to the ``num_targets``
      target values; ``epsilon`` regularization parameter. Remaining ``kwargs``
      are passed on to parameterize the
      :class:`~ott.solvers.linear.sinkhorn.Sinkhorn` solver.

  Returns:
    An Array of the same shape as the input with soft-rank values
    normalized to be in :math:`[0, n-1]` where :math:`n` is
    ``inputs.shape[axis]``.
  """
  return apply_on_axis(
      _ranks, inputs, axis, num_targets, target_weights, **kwargs
  )


def topk_mask(
    inputs: jnp.ndarray,
    axis: int = -1,
    k: int = 1,
    **kwargs: Any,
) -> jnp.ndarray:
  r"""Soft :math:`\text{top-}k` selection mask.

  For instance:

  .. code-block:: python

    k = 5
    x = jax.random.uniform(rng, (100,))
    mask = topk_mask(x, k=k)

  will output a vector of shape ``x.shape``, with values in :math:`[0,1]`, that
  are differentiable approximations to the binary mask selecting the top $k$
  entries in ``x``. These should be compared to the non-differentiable mask
  obtained with :func:`jax.numpy.sort`, which can be obtained as:

  .. code-block:: python

    mask = x >= jax.numpy.sort(x).flip()[k-1]

  Args:
    inputs: Array of any shape.
    axis: the axis on which to apply the soft-sorting operator.
    k: topk parameter. Should be smaller than ``inputs.shape[axis]``.
    kwargs: keyword arguments passed on to lower level functions. Of interest
      to the user are ``squashing_fun``, which will redistribute the values in
      ``inputs`` to lie in :math:`[0,1]` (sigmoid of whitened values by default)
      to solve the optimal transport problem;
      :class:`cost_fn <ott.geometry.costs.CostFn>` object of
      :class:`~ott.geometry.pointcloud.PointCloud`, which defines the ground
      1D cost function to transport from ``inputs`` to the ``num_targets``
      target values; ``epsilon`` regularization parameter. Remaining ``kwargs``
      are passed on to parameterize the
      :class:`~ott.solvers.linear.sinkhorn.Sinkhorn` solver.

  Returns:
    The soft :math:`\text{top-}k` selection mask.
  """
  num_points = inputs.shape[axis]
  assert k < num_points, (
      f"`k` must be smaller than `inputs.shape[axis]`, yet {k} >= {num_points}."
  )
  target_weights = jnp.array([1.0 - k / num_points, k / num_points])
  out = apply_on_axis(
      _ranks,
      inputs,
      axis,
      num_targets=None,
      target_weights=target_weights,
      **kwargs
  )
  return out / (num_points - 1)


def quantile(
    inputs: jnp.ndarray,
    q: Optional[Union[float, jnp.ndarray]],
    axis: Union[int, Tuple[int, ...]] = -1,
    weight: Optional[Union[float, jnp.ndarray]] = None,
    **kwargs: Any,
) -> jnp.ndarray:
  r"""Apply the soft quantiles operator on the input tensor.

  For instance:

  .. code-block:: python

    x = jax.random.uniform(rng, (100,))
    x_quantiles = quantile(x, q=jnp.array([0.2, 0.8]))

  ``x_quantiles`` will hold an approximation to the 20 and 80 percentiles in
  ``x``, computed as a convex combination (a weighted mean, with weights summing
  to 1) of all values in ``x`` (and not, as for standard quantiles, the
  values ``x_sorted[20]`` and ``x_sorted[80]`` if ``x_sorted=jnp.sort(x)``).
  These values offer a trade-off between accuracy (closeness to the true
  percentiles) and gradient (the Jacobian of ``x_quantiles`` w.r.t ``x`` will
  impact all values listed in ``x``, not just those indexed at 20 and 80).

  The non-differentiable version is given by :func:`jax.numpy.quantile`, e.g.

  .. code-block:: python

    x_quantiles = jax.numpy.quantile(x, q=jnp.array([0.2, 0.8]))

  Args:
   inputs: an Array of any shape.
   q: values of the quantile level to be computed, e.g. [0.5] for median.
     These values should all lie within the segment :math:`]0,1[`, excluding
     boundary values :math:`0` and :math:`1`.
   axis: the axis on which to apply the operator.
   weight: the weight assigned to each quantile target value in the OT problem.
     This weight should be small, typically of the order of ``1/n``, where ``n``
     is the size of ``x``. Note: Since the size of ``q`` times ``weight``
     must be strictly smaller than ``1``, in order to leave enough mass to set
     other target values in the transport problem, the algorithm might ensure
     this by setting, when needed, a lower value.
   kwargs: keyword arguments passed on to lower level functions. Of interest
     to the user are ``squashing_fun``, which will redistribute the values in
     ``inputs`` to lie in :math:`[0,1]` (sigmoid of whitened values by default)
     to solve the optimal transport problem;
     :class:`cost_fn <ott.geometry.costs.CostFn>` object of
     :class:`~ott.geometry.pointcloud.PointCloud`, which defines the ground
     1D cost function to transport from ``inputs`` to the ``num_targets``
     target values; ``epsilon`` regularization parameter. Remaining ``kwargs``
     are passed on to parameterize the
     :class:`~ott.solvers.linear.sinkhorn.Sinkhorn` solver.

  Returns:
    An Array, which has the same shape as ``inputs``, except on the ``axis``
    that is passed, which has size ``q.shape[0]``, to collect soft-quantile
    values.
  """

  def _quantile(
      inputs: jnp.ndarray, q: float, weight: float, **kwargs
  ) -> jnp.ndarray:
    num_points = inputs.shape[0]
    q = jnp.array([0.2, 0.5, 0.8]) if q is None else jnp.atleast_1d(q)
    num_quantiles = q.shape[0]
    a = jnp.ones((num_points,)) / num_points
    idx = jnp.argsort(q)
    q = q[idx]

    extended_q = jnp.concatenate([jnp.array([0.0]), q, jnp.array([1.0])])
    filler_weights = extended_q[1:] - extended_q[:-1]
    safe_weight = 0.5 * jnp.concatenate([
        jnp.array([1.0 / num_quantiles]), filler_weights
    ])
    if weight is None:
      # Populate with other options.
      safe_weight = jnp.concatenate([
          safe_weight,
          jnp.array(
              [0.02]
          ),  # reasonable mass per quantile for a small number of points
          jnp.array(
              [1.5 / num_points]
          ),  # this means each quantile would be ~ assigned 1.5 points.
      ])
    else:
      safe_weight = jnp.concatenate([safe_weight, jnp.atleast_1d(weight)])
    weight = jnp.min(safe_weight)
    weights = jnp.ones(filler_weights.shape) * weight

    # Takes into account quantile_width in the definition of weights
    shift = -jnp.ones(filler_weights.shape)
    shift = shift + 0.5 * (
        jax.nn.one_hot(0, num_quantiles + 1) +
        jax.nn.one_hot(num_quantiles, num_quantiles + 1)
    )
    filler_weights = filler_weights + weights * shift

    # Adds one more value to have tensors of the same shape to interleave them.
    quantile_weights = jnp.ones(num_quantiles + 1) * weights

    # Interleaves the filler_weights with the quantile weights.
    weights = jnp.reshape(
        jnp.stack([filler_weights, quantile_weights], axis=1), (-1,)
    )[:-1]

    ot = transport_for_sort(inputs, a, weights, **kwargs)
    out = 1.0 / weights * ot.apply(jnp.squeeze(inputs), axis=0)

    # Recover odd indices corresponding to the desired quantiles.
    odds = np.concatenate([
        np.zeros((num_quantiles + 1, 1), dtype=bool),
        np.ones((num_quantiles + 1, 1), dtype=bool)
    ],
                          axis=1).ravel()[:-1]
    return out[odds][idx]

  return apply_on_axis(_quantile, inputs, axis, q, weight, **kwargs)


def multivariate_cdf_quantile_maps(
    inputs: jnp.ndarray,
    target_sampler: Optional[Callable[[jax.Array, Tuple[int, int]],
                                      jax.Array]] = None,
    rng: Optional[jax.Array] = None,
    num_target_samples: Optional[int] = None,
    cost_fn: Optional[costs.CostFn] = None,
    epsilon: Optional[float] = None,
    input_weights: Optional[jnp.ndarray] = None,
    target_weights: Optional[jnp.ndarray] = None,
    **kwargs: Any
) -> Tuple[Func_t, Func_t]:
  r"""Returns multivariate CDF and quantile maps, given input samples.

  Implements the multivariate generalizations for CDF and quantiles proposed in
  :cite:`chernozhukov:17`. The reference measure is assumed to be the uniform
  measure by default, but can be modified. For consistency, the reference
  measure should be symmetrically centered around
  :math:`(\tfrac{1}{2},\cdots,\tfrac{1}{2})` and supported on :math:`[0, 1]^d`.

  The implementation return two entropic map estimators, one for the CDF map,
  the other for the quantiles map.

  Args:
    inputs: 2D array of ``[n, d]`` vectors.
    target_sampler: Callable that takes a ``rng`` and ``[m, d]`` shape.
      ``m`` is passed on as ``target_num_samples``, dimension ``d`` is inferred
      directly from the shape passed in ``inputs``. This is assumed by default
      to be :func:`~jax.random.uniform`, and could be any other random sampler
      properly wrapped to have the signature above.
    rng: rng key used by ``target_sampler``.
    num_target_samples: number ``m`` of points generated in the target
      distribution.
    cost_fn: Cost function, used to compare ``inputs`` and ``targets``.
      Passed on to instantiate a
      :class:`~ott.geometry.pointcloud.PointCloud` object. If :obj:`None`,
      :class:`~ott.geometry.costs.SqEuclidean` is used.
    epsilon: entropic regularization parameter used to instantiate the
      :class:`~ott.geometry.pointcloud.PointCloud` object.
    input_weights: ``[n,]`` vector of weights for input measure. Assumed to
      be uniform by default.
    target_weights: ``[m,]`` vector of weights for target measure. Assumed
      to be uniform by default.
    kwargs: keyword arguments passed on to the :func:`~ott.solvers.linear.solve`
      function, which solves the OT problem between ``inputs`` and ``targets``
      using the :class:`~ott.solvers.linear.sinkhorn.Sinkhorn` algorithm.

  Returns:
    - The multivariate CDF map, taking a ``[b, d]`` batch of vectors in the
      range of the ``inputs``, and mapping each vector within the range
      of the reference measure (assumed by default to be :math:`[0, 1]^d`).
    - The quantile map, mapping a batch ``[b, d]`` of multivariate quantile
      vectors onto ``[b, d]`` vectors in :math:`[0, 1]^d`, the range of
      the reference measure.
  """
  n, d = inputs.shape
  rng = utils.default_prng_key(rng)

  if num_target_samples is None:
    num_target_samples = n
  if target_sampler is None:
    target_sampler = jax.random.uniform

  targets = target_sampler(rng, (num_target_samples, d))
  geom = pointcloud.PointCloud(
      inputs, targets, cost_fn=cost_fn, epsilon=epsilon
  )

  out = linear.solve(geom, a=input_weights, b=target_weights, **kwargs)
  potentials = out.to_dual_potentials()

  cdf_map = jtu.Partial(lambda x, p: p.transport(x), p=potentials)
  quantile_map = jtu.Partial(
      lambda x, p: p.transport(x, forward=False), p=potentials
  )
  return cdf_map, quantile_map


def _quantile_normalization(
    inputs: jnp.ndarray, targets: jnp.ndarray, weights: float, **kwargs: Any
) -> jnp.ndarray:
  """Apply soft quantile normalization on a one dimensional array."""
  num_points = inputs.shape[0]
  a = jnp.ones((num_points,)) / num_points
  ot = transport_for_sort(inputs, a, weights, **kwargs)
  return 1.0 / a * ot.apply(targets, axis=1)


def quantile_normalization(
    inputs: jnp.ndarray,
    targets: jnp.ndarray,
    weights: Optional[jnp.ndarray] = None,
    axis: int = -1,
    **kwargs: Any,
) -> jnp.ndarray:
  r"""Re-normalize inputs so that its quantiles match those of targets/weights.

  Quantile normalization rearranges the values in inputs to values that match
  the distribution of values described in the discrete distribution ``targets``
  weighted by ``weights``. This transformation preserves the order of values
  in ``inputs`` along the specified ``axis``.

  Args:
    inputs: array of any shape whose values will be changed to match those in
      ``targets``.
    targets: sorted array (in ascending order) of dimension 1 describing a
      discrete distribution. Note: the ``targets`` values must be provided as
      a sorted vector.
    weights: vector of non-negative weights, summing to :math:`1`, of the same
      size as ``targets``. When not set, this defaults to the uniform
      distribution.
    axis: the axis along which the quantile transformation is applied.
    kwargs: keyword arguments passed on to lower level functions. Of interest
      to the user are ``squashing_fun``, which will redistribute the values in
      ``inputs`` to lie in :math:`[0,1]` (sigmoid of whitened values by default)
      to solve the optimal transport problem;
      :class:`cost_fn <ott.geometry.costs.CostFn>` object of
      :class:`~ott.geometry.pointcloud.PointCloud`, which defines the ground
      1D cost function to transport from ``inputs`` to the ``num_targets``
      target values; ``epsilon`` regularization parameter. Remaining ``kwargs``
      are passed on to parameterize the
      :class:`~ott.solvers.linear.sinkhorn.Sinkhorn` solver.

  Returns:
    An array, which has the same shape as the input, except on the give axis on
    which the dimension is 1.

  Raises:
    A ValueError in case the weights and the targets are both set and not of
    compatible shapes.
  """
  if weights is not None and weights.shape != targets.shape:
    raise ValueError(
        "The target weights and targets values should have the "
        f"same shape: {targets.shape} != {weights.shape}"
    )
  if weights is None:
    num_targets = targets.shape[0]
    weights = jnp.ones((num_targets,)) / num_targets

  op = _quantile_normalization
  return apply_on_axis(op, inputs, axis, targets, weights, **kwargs)


def sort_with(
    inputs: jnp.ndarray,
    criterion: jnp.ndarray,
    topk: int = -1,
    **kwargs: Any,
) -> jnp.ndarray:
  r"""Sort a multidimensional array according to a real valued criterion.

  Given ``batch`` vectors of dimension `dim`, to which, for each, a real value
  ``criterion`` is associated, this function produces ``topk`` (or
  ``batch`` if ``topk`` is set to -1, as by default) composite vectors of size
  ``dim`` that will be convex combinations of all vectors, ranked from smallest
  to largest criterion. Composite vectors with the largest indices will contain
  convex combinations of those vectors with highest criterion, vectors with
  smaller indices will contain combinations of vectors with smaller criterion.

  Args:
    inputs: Array of size [batch, dim].
    criterion: the values according to which to sort the inputs. It has shape
      [batch, 1].
    topk: The number of outputs to keep.
    kwargs: keyword arguments passed on to lower level functions. Of interest
      to the user are ``squashing_fun``, which will redistribute the values in
      ``inputs`` to lie in :math:`[0,1]` (sigmoid of whitened values by default)
      to solve the optimal transport problem;
      :class:`cost_fn <ott.geometry.costs.CostFn>` object of
      :class:`~ott.geometry.pointcloud.PointCloud`, which defines the ground
      1D cost function to transport from ``inputs`` to the ``num_targets``
      target values; ``epsilon`` regularization parameter. Remaining ``kwargs``
      are passed on to parameterize the
      :class:`~ott.solvers.linear.sinkhorn.Sinkhorn` solver.

  Returns:
    An Array of size [batch | topk, dim].
  """
  num_points = criterion.shape[0]
  weights = jnp.ones(num_points) / num_points
  if 0 < topk < num_points:
    start_index = 1
    target_weights = jnp.concatenate([
        jnp.array([(num_points - topk) / num_points]),
        jnp.ones(topk) / num_points
    ])
  else:
    start_index = 0
    target_weights = jnp.ones((num_points,)) / num_points
  ot = transport_for_sort(criterion, weights, target_weights, **kwargs)
  # Applies the topk on each of the dimensions of the inputs.
  sort_fn = jax.vmap(
      lambda x: (1.0 / target_weights * ot.apply(x, axis=0))[start_index:],
      in_axes=(1,),
      out_axes=1
  )
  return sort_fn(inputs)


def _quantize(inputs: jnp.ndarray, num_q: int, **kwargs: Any) -> jnp.ndarray:
  """Apply the soft quantization operator on a one dimensional array."""
  num_points = inputs.shape[0]
  a = jnp.ones((num_points,)) / num_points
  b = jnp.ones((num_q,)) / num_q
  ot = transport_for_sort(inputs, a, b, **kwargs)
  return 1.0 / a * ot.apply(1.0 / b * ot.apply(inputs), axis=1)


def quantize(
    inputs: jnp.ndarray,
    num_levels: int = 10,
    axis: int = -1,
    **kwargs: Any,
) -> jnp.ndarray:
  r"""Soft quantizes an input according using ``num_levels`` values along axis.

  The quantization operator consists in concentrating several values around
  a few predefined ``num_levels``. The soft quantization operator proposed here
  does so by carrying out a `soft` concentration that is differentiable. The
  ``inputs`` values are first soft-sorted, resulting in ``num_levels`` values.
  In a second step, the ``inputs`` values are then provided again a composite
  value that is equal (for each) to a convex combination of those soft-sorted
  values using the transportation matrix. As the regularization parameter
  ``epsilon`` of regularized optimal transport goes to 0, this operator recovers
  the expected behavior of quantization, namely each value in ``inputs`` is
  assigned a single level. When using ``epsilon>0`` the behavior is similar but
  differentiable.

  Args:
    inputs: an Array of size [batch, dim].
    num_levels: number of quantiles available to quantize the signal.
    axis: axis along which quantization is carried out.
    kwargs: keyword arguments passed on to lower level functions. Of interest
      to the user are ``squashing_fun``, which will redistribute the values in
      ``inputs`` to lie in :math:`[0,1]` (sigmoid of whitened values by default)
      to solve the optimal transport problem;
      :class:`cost_fn <ott.geometry.costs.CostFn>` object of
      :class:`~ott.geometry.pointcloud.PointCloud`, which defines the ground
      1D cost function to transport from ``inputs`` to the ``num_targets``
      target values; ``epsilon`` regularization parameter. Remaining ``kwargs``
      are passed on to parameterize the
      :class:`~ott.solvers.linear.sinkhorn.Sinkhorn` solver.

  Returns:
    An Array of the same size as ``inputs``.
  """
  return apply_on_axis(_quantize, inputs, axis, num_levels, **kwargs)



# Copyright OTT-JAX
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any, Literal, Optional, Tuple, Union

import jax
import jax.numpy as jnp

from ott.geometry import costs, pointcloud
from ott.solvers import linear, quadratic

__all__ = [
    "match_linear",
    "match_quadratic",
    "sample_joint",
    "sample_conditional",
    "uniform_sampler",
]

ScaleCost_t = Union[float, Literal["mean", "max_cost", "median"]]


def match_linear(
    x: jnp.ndarray,
    y: Optional[jnp.ndarray],
    cost_fn: Optional[costs.CostFn] = None,
    epsilon: Optional[float] = None,
    scale_cost: ScaleCost_t = 1.0,
    **kwargs: Any
) -> jnp.ndarray:
  """Compute solution to a linear OT problem.

  Args:
    x: Source point cloud of shape ``[n, d]``.
    y: Target point cloud of shape ``[m, d]``.
    cost_fn: Cost function.
    epsilon: Regularization parameter.
    scale_cost: Scaling of the cost matrix.
    kwargs: Additional arguments for :func:`ott.solvers.linear.solve`.

  Returns:
    Optimal transport matrix.
  """
  geom = pointcloud.PointCloud(
      x, y, cost_fn=cost_fn, epsilon=epsilon, scale_cost=scale_cost
  )
  out = linear.solve(geom, **kwargs)
  return out.matrix


def match_quadratic(
    xx: jnp.ndarray,
    yy: jnp.ndarray,
    x: Optional[jnp.ndarray] = None,
    y: Optional[jnp.ndarray] = None,
    scale_cost: ScaleCost_t = 1.0,
    cost_fn: Optional[costs.CostFn] = None,
    **kwargs: Any
) -> jnp.ndarray:
  """Compute solution to a quadratic OT problem.

  Args:
    xx: Source point cloud of shape ``[n, d1]``.
    yy: Target point cloud of shape ``[m, d2]``.
    x: Linear (fused) term of the source point cloud.
    y: Linear (fused) term of the target point cloud.
    scale_cost: Scaling of the cost matrix.
    cost_fn: Cost function.
    kwargs: Additional arguments for :func:`ott.solvers.quadratic.solve`.

  Returns:
    Optimal transport matrix.
  """
  geom_xx = pointcloud.PointCloud(xx, cost_fn=cost_fn, scale_cost=scale_cost)
  geom_yy = pointcloud.PointCloud(yy, cost_fn=cost_fn, scale_cost=scale_cost)
  if x is None:
    geom_xy = None
  else:
    geom_xy = pointcloud.PointCloud(
        x, y, cost_fn=cost_fn, scale_cost=scale_cost
    )

  out = quadratic.solve(geom_xx, geom_yy, geom_xy, **kwargs)
  return out.matrix


def sample_joint(rng: jax.Array,
                 tmat: jnp.ndarray) -> Tuple[jnp.ndarray, jnp.ndarray]:
  """Sample jointly from a transport matrix.

  Args:
    rng: Random number generator.
    tmat: Transport matrix of shape ``[n, m]``.

  Returns:
    Source and target indices of shape ``[n,]`` and ``[m,]``, respectively.
  """
  n, m = tmat.shape
  tmat_flattened = tmat.flatten()
  indices = jax.random.choice(
      rng, len(tmat_flattened), p=tmat_flattened, shape=[n]
  )
  src_ixs = indices // m
  tgt_ixs = indices % m
  return src_ixs, tgt_ixs


def sample_conditional(
    rng: jax.Array,
    tmat: jnp.ndarray,
    *,
    k: int = 1,
) -> Tuple[jnp.ndarray, jnp.ndarray]:
  """Sample conditionally from a transport matrix.

  Args:
    rng: Random number generator.
    tmat: Transport matrix of shape ``[n, m]``.
    k: Expected number of samples to sample per source sample.

  Returns:
    Source and target indices of shape ``[n, k]`` and ``[m, k]``, respectively.
  """
  assert k > 0, "Number of samples per source must be positive."
  n, m = tmat.shape

  src_marginals = tmat.sum(axis=1)
  rng, rng_ixs = jax.random.split(rng, 2)
  indices = jax.random.choice(rng_ixs, a=n, p=src_marginals, shape=(n,))
  tmat = tmat[indices]

  rngs = jax.random.split(rng, n)
  tgt_ixs = jax.vmap(
      lambda rng, row: jax.random.choice(rng, a=m, p=row, shape=(k,)),
      in_axes=[0, 0],
  )(rngs, tmat)  # (m, k)

  src_ixs = jnp.repeat(indices[:, None], k, axis=1)  # (n, k)
  return src_ixs, tgt_ixs


def uniform_sampler(
    rng: jax.Array,
    num_samples: int,
    low: float = 0.0,
    high: float = 1.0,
    offset: Optional[float] = None
) -> jnp.ndarray:
  r"""Sample from a uniform distribution.

  Sample :math:`t` from a uniform distribution :math:`[low, high]`.
  If `offset` is not :obj:`None`, one element :math:`t` is sampled from
  :math:`[low, high]` and the K samples are constructed via
  :math:`(t + k)/K \mod (high - low - offset) + low`.

  Args:
    rng: Random number generator.
    num_samples: Number of samples to generate.
    low: Lower bound of the uniform distribution.
    high: Upper bound of the uniform distribution.
    offset: Offset of the uniform distribution.
      If :obj:`None`, no offset is used.

  Returns:
    An array of shape ``[num_samples, 1]``.
  """
  if offset is None:
    return jax.random.uniform(rng, (num_samples, 1), minval=low, maxval=high)

  t = jax.random.uniform(rng, (1, 1), minval=low, maxval=high)
  mod_term = ((high - low) - offset)
  return (t + jnp.arange(num_samples)[:, None] / num_samples) % mod_term



# Copyright OTT-JAX
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any, Callable, Literal, Optional, Tuple, Union

import jax
import jax.numpy as jnp

from ott import utils
from ott.geometry import costs, geometry, low_rank
from ott.math import utils as mu

__all__ = ["PointCloud"]


@jax.tree_util.register_pytree_node_class
class PointCloud(geometry.Geometry):
  """Defines geometry for 2 point clouds (possibly 1 vs itself).

  Creates a geometry, specifying a cost function passed as CostFn type object.
  When the number of points is large, setting the ``batch_size`` flag implies
  that cost and kernel matrices used to update potentials or scalings
  will be recomputed on the fly, rather than stored in memory. More precisely,
  when setting ``batch_size``, the cost function will be partially cached by
  storing norm values for each point in both point clouds, but the pairwise cost
  function evaluations won't be.

  Args:
    x : n x d array of n d-dimensional vectors
    y : m x d array of m d-dimensional vectors. If `None`, use ``x``.
    cost_fn: a CostFn function between two points in dimension d.
    batch_size: When ``None``, the cost matrix corresponding to that point cloud
     is computed, stored and later re-used at each application of
     :meth:`apply_lse_kernel`. When ``batch_size`` is a positive integer,
     computations are done in an online fashion, namely the cost matrix is
     recomputed at each call of the :meth:`apply_lse_kernel` step,
     ``batch_size`` lines at a time, used on a vector and discarded.
     The online computation is particularly useful for big point clouds
     whose cost matrix does not fit in memory.
    scale_cost: option to rescale the cost matrix. Implemented scalings are
      'median', 'mean', 'max_cost', 'max_norm' and 'max_bound'.
      Alternatively, a float factor can be given to rescale the cost such
      that ``cost_matrix /= scale_cost``.
    kwargs: keyword arguments for :class:`~ott.geometry.geometry.Geometry`.
  """

  def __init__(
      self,
      x: jnp.ndarray,
      y: Optional[jnp.ndarray] = None,
      cost_fn: Optional[costs.CostFn] = None,
      batch_size: Optional[int] = None,
      scale_cost: Union[float, Literal["mean", "max_norm", "max_bound",
                                       "max_cost", "median"]] = 1.0,
      **kwargs: Any,
  ):
    super().__init__(**kwargs)
    self.x = x
    self.y = self.x if y is None else y

    self.cost_fn = costs.SqEuclidean() if cost_fn is None else cost_fn
    if batch_size is not None:
      assert batch_size > 0, f"`batch_size={batch_size}` must be positive."
    self._batch_size = batch_size
    self._scale_cost = scale_cost

  def apply_lse_kernel(  # noqa: D102
      self,
      f: jnp.ndarray,
      g: jnp.ndarray,
      eps: float,
      vec: Optional[jnp.ndarray] = None,
      axis: int = 0
  ) -> Tuple[jnp.ndarray, jnp.ndarray]:
    if not self.is_online:
      return super().apply_lse_kernel(f, g, eps, vec, axis)

    def apply(x: jnp.ndarray, y: jnp.ndarray, f: jnp.ndarray,
              g: jnp.ndarray) -> Tuple[jnp.ndarray, jnp.ndarray]:
      x, y = jnp.atleast_2d(x), jnp.atleast_2d(y)
      cost = self.cost_fn.all_pairs(x, y) * inv_scale_cost
      cost = cost.squeeze(1 - axis)
      # axis=-1
      res, sgn = mu.logsumexp((f + g - cost) / eps, b=vec, return_sign=True)
      return eps * res, sgn

    inv_scale_cost = self.inv_scale_cost
    in_axes = (None, 0, None, 0) if axis == 0 else (0, None, 0, None)
    batched_apply = utils.batched_vmap(
        apply,
        batch_size=self.batch_size,
        in_axes=in_axes,
    )
    w_res, w_sgn = batched_apply(self.x, self.y, f, g)
    remove = f if axis == 1 else g
    return w_res - jnp.where(jnp.isfinite(remove), remove, 0), w_sgn

  def apply_kernel(  # noqa: D102
      self,
      vec: jnp.ndarray,
      eps: Optional[float] = None,
      axis: int = 0
  ) -> jnp.ndarray:
    if eps is None:
      eps = self.epsilon
    if not self.is_online:
      return super().apply_kernel(vec, eps, axis)

    def apply(x: jnp.ndarray, y: jnp.ndarray, vec: jnp.ndarray) -> jnp.ndarray:
      x, y = jnp.atleast_2d(x), jnp.atleast_2d(y)
      cost = self.cost_fn.all_pairs(x, y) * inv_scale_cost
      cost = cost.squeeze(1 - axis)
      return jnp.dot(jnp.exp(-cost / eps), vec)

    inv_scale_cost = self.inv_scale_cost
    in_axes = (None, 0, None) if axis == 0 else (0, None, None)
    batched_apply = utils.batched_vmap(
        apply, batch_size=self.batch_size, in_axes=in_axes
    )
    return batched_apply(self.x, self.y, vec)

  def _apply_cost_to_vec(
      self,
      vec: jnp.ndarray,
      axis: int = 0,
      fn: Optional[Callable[[jnp.ndarray], jnp.ndarray]] = None,
      is_linear: bool = False,
      scale_cost: Optional[float] = None,
  ) -> jnp.ndarray:

    def apply(x: jnp.ndarray, y: jnp.ndarray, arr: jnp.ndarray) -> jnp.ndarray:
      x, y = jnp.atleast_2d(x), jnp.atleast_2d(y)
      cost = self.cost_fn.all_pairs(x, y) * scale_cost
      cost = cost.squeeze(1 - axis)
      if fn is not None:
        cost = fn(cost)
      return jnp.dot(cost, arr)

    # when computing the online properties, this is set to 1.0
    if scale_cost is None:
      scale_cost = self.inv_scale_cost
    # switch to an efficient computation for the squared Euclidean case
    if self.is_squared_euclidean and (fn is None or is_linear):
      return self._apply_sqeucl_cost(
          vec,
          scale_cost,
          axis=axis,
          fn=fn,
      )

    # materialize the cost
    if not self.is_online:
      return super()._apply_cost_to_vec(
          vec, axis=axis, fn=fn, is_linear=is_linear
      )

    in_axes = (None, 0, None) if axis == 0 else (0, None, None)
    batched_apply = utils.batched_vmap(
        apply, batch_size=self.batch_size, in_axes=in_axes
    )
    return batched_apply(self.x, self.y, vec)

  def _apply_sqeucl_cost(
      self,
      vec: jnp.ndarray,
      scale_cost: float,
      axis: int = 0,
      fn: Optional[Callable[[jnp.ndarray], jnp.ndarray]] = None,
  ) -> jnp.ndarray:
    assert vec.ndim == 1, vec.shape
    assert self.is_squared_euclidean, "Cost matrix is not a squared Euclidean."
    x, y = (self.x, self.y) if axis == 0 else (self.y, self.x)
    nx, ny = self.cost_fn.norm(x), self.cost_fn.norm(y)

    applied_cost = jnp.dot(nx, vec) + ny * jnp.sum(vec, axis=0)
    applied_cost = applied_cost - 2.0 * jnp.dot(y, jnp.dot(x.T, vec))
    if fn is not None:
      applied_cost = fn(applied_cost)
    return scale_cost * applied_cost

  def _compute_summary_online(
      self, summary: Literal["mean", "max_cost"]
  ) -> jnp.ndarray:
    """Compute mean or max of cost matrix online, i.e. without instantiating it.

    Args:
      summary: can be 'mean' or 'max_cost'.

    Returns:
      summary statistics
    """

    def compute_max(x: jnp.ndarray, y: jnp.ndarray) -> jnp.ndarray:
      x, y = jnp.atleast_2d(x), jnp.atleast_2d(y)
      cost = self.cost_fn.all_pairs(x, y)
      return jnp.max(jnp.abs(cost))

    if summary == "mean":
      a, b = self._n_normed_ones, self._m_normed_ones
      return jnp.sum(self._apply_cost_to_vec(a, scale_cost=1.0) * b)

    if summary == "max_cost":
      fn = utils.batched_vmap(
          compute_max, batch_size=self.batch_size, in_axes=[0, None]
      )
      return jnp.max(fn(self.x, self.y))

    raise ValueError(
        f"Scaling method {summary} does not exist for online mode."
    )

  def barycenter(self, weights: jnp.ndarray) -> jnp.ndarray:
    """Compute barycenter of points in self.x using weights."""
    return self.cost_fn.barycenter(self.x, weights)[0]

  @classmethod
  def prepare_divergences(
      cls,
      x: jnp.ndarray,
      y: jnp.ndarray,
      static_b: bool = False,
      src_mask: Optional[jnp.ndarray] = None,
      tgt_mask: Optional[jnp.ndarray] = None,
      **kwargs: Any
  ) -> Tuple["PointCloud", ...]:
    """Instantiate the geometries used for a divergence computation."""
    couples = [(x, y), (x, x)]
    masks = [(src_mask, tgt_mask), (src_mask, src_mask)]
    if not static_b:
      couples += [(y, y)]
      masks += [(tgt_mask, tgt_mask)]

    return tuple(
        cls(x, y, src_mask=x_mask, tgt_mask=y_mask, **kwargs)
        for ((x, y), (x_mask, y_mask)) in zip(couples, masks)
    )

  def tree_flatten(self):  # noqa: D102
    return (
        self.x,
        self.y,
        self._src_mask,
        self._tgt_mask,
        self._epsilon_init,
        self.cost_fn,
    ), {
        "batch_size": self._batch_size,
        "scale_cost": self._scale_cost
    }

  @classmethod
  def tree_unflatten(cls, aux_data, children):  # noqa: D102
    x, y, src_mask, tgt_mask, epsilon, cost_fn = children
    return cls(
        x,
        y,
        cost_fn=cost_fn,
        src_mask=src_mask,
        tgt_mask=tgt_mask,
        epsilon=epsilon,
        **aux_data
    )

  def _cosine_to_sqeucl(self) -> "PointCloud":
    assert isinstance(self.cost_fn, costs.Cosine), type(self.cost_fn)
    (x, y, *args, _), aux_data = self.tree_flatten()
    x = x / jnp.linalg.norm(x, axis=-1, keepdims=True)
    y = y / jnp.linalg.norm(y, axis=-1, keepdims=True)
    # TODO(michalk8): find a better way
    aux_data["scale_cost"] = 2.0 / self.inv_scale_cost
    cost_fn = costs.SqEuclidean()
    return type(self).tree_unflatten(aux_data, [x, y] + args + [cost_fn])

  def to_LRCGeometry(
      self,
      scale: float = 1.0,
      **kwargs: Any,
  ) -> Union[low_rank.LRCGeometry, "PointCloud"]:
    r"""Convert point cloud to low-rank geometry.

    Args:
      scale: Value used to rescale the factors of the low-rank geometry.
        Useful when this geometry is used in the linear term of fused GW.
      kwargs: Keyword arguments, such as ``rank``, to
        :meth:`~ott.geometry.geometry.Geometry.to_LRCGeometry` used when
        the point cloud does not have squared Euclidean cost.

    Returns:
      Returns the unmodified point cloud if :math:`n m \ge (n + m) d`, where
      :math:`n, m` is the shape and :math:`d` is the dimension of the point
      cloud with squared Euclidean cost.
      Otherwise, returns the re-scaled low-rank geometry.
    """
    if self.is_squared_euclidean:
      if self._check_LRC_dim:
        return self._sqeucl_to_lr(scale)
      # we don't update the `scale_factor` because in GW, the linear cost
      # is first materialized and then scaled by `fused_penalty` afterwards
      return self
    return super().to_LRCGeometry(scale=scale, **kwargs)

  def _sqeucl_to_lr(self, scale: float = 1.0) -> low_rank.LRCGeometry:
    assert self.is_squared_euclidean, "Geometry must be squared Euclidean."
    n, m = self.shape
    nx = jnp.sum(self.x ** 2, axis=1, keepdims=True)
    ny = jnp.sum(self.y ** 2, axis=1, keepdims=True)
    cost_1 = jnp.concatenate(
        (nx, jnp.ones((n, 1), dtype=self.dtype), -(2.0 ** 0.5) * self.x),
        axis=1,
    )
    cost_2 = jnp.concatenate(
        (jnp.ones((m, 1), dtype=self.dtype), ny, (2.0 ** 0.5) * self.y),
        axis=1,
    )

    return low_rank.LRCGeometry(
        cost_1=cost_1,
        cost_2=cost_2,
        scale_factor=scale,
        epsilon=self._epsilon_init,
        relative_epsilon=self._relative_epsilon,
        scale_cost=self._scale_cost,
        src_mask=self.src_mask,
        tgt_mask=self.tgt_mask,
    )

  def subset(  # noqa: D102
      self, src_ixs: Optional[jnp.ndarray], tgt_ixs: Optional[jnp.ndarray],
      **kwargs: Any
  ) -> "PointCloud":

    def subset_fn(
        arr: Optional[jnp.ndarray],
        ixs: Optional[jnp.ndarray],
    ) -> jnp.ndarray:
      return arr if arr is None or ixs is None else arr[ixs, ...]

    return self._mask_subset_helper(
        src_ixs, tgt_ixs, fn=subset_fn, propagate_mask=True, **kwargs
    )

  def mask(  # noqa: D102
      self,
      src_mask: Optional[jnp.ndarray],
      tgt_mask: Optional[jnp.ndarray],
      mask_value: float = 0.0,
  ) -> "PointCloud":

    def mask_fn(
        arr: Optional[jnp.ndarray],
        mask: Optional[jnp.ndarray],
    ) -> Optional[jnp.ndarray]:
      if arr is None or mask is None:
        return arr
      return jnp.where(mask[:, None], arr, mask_value)

    src_mask = self._normalize_mask(src_mask, self.shape[0])
    tgt_mask = self._normalize_mask(tgt_mask, self.shape[1])
    return self._mask_subset_helper(
        src_mask, tgt_mask, fn=mask_fn, propagate_mask=False
    )

  def _mask_subset_helper(
      self,
      src_ixs: Optional[jnp.ndarray],
      tgt_ixs: Optional[jnp.ndarray],
      *,
      fn: Callable[[Optional[jnp.ndarray], Optional[jnp.ndarray]],
                   Optional[jnp.ndarray]],
      propagate_mask: bool,
      **kwargs: Any,
  ) -> "PointCloud":
    (x, y, src_mask, tgt_mask, *children), aux_data = self.tree_flatten()
    x = fn(x, src_ixs)
    y = fn(y, tgt_ixs)
    if propagate_mask:
      src_mask = self._normalize_mask(src_mask, self.shape[0])
      tgt_mask = self._normalize_mask(tgt_mask, self.shape[1])
      src_mask = fn(src_mask, src_ixs)
      tgt_mask = fn(tgt_mask, tgt_ixs)
    aux_data = {**aux_data, **kwargs}

    return type(self).tree_unflatten(
        aux_data, [x, y, src_mask, tgt_mask] + children
    )

  @property
  def cost_matrix(self) -> Optional[jnp.ndarray]:  # noqa: D102
    return self.inv_scale_cost * self._unscaled_cost_matrix

  @property
  def _unscaled_cost_matrix(self) -> jnp.ndarray:
    return self.cost_fn.all_pairs(self.x, self.y)

  @property
  def inv_scale_cost(self) -> float:  # noqa: D102
    if isinstance(self._scale_cost, (int, float, jax.Array)):
      return 1.0 / self._scale_cost
    self = self._masked_geom()
    if self._scale_cost == "max_cost":
      if self.is_online:
        return 1.0 / self._compute_summary_online(self._scale_cost)
      return 1.0 / jnp.max(self._unscaled_cost_matrix)
    if self._scale_cost == "mean":
      if self.is_online:
        return 1.0 / self._compute_summary_online(self._scale_cost)
      geom = self._masked_geom(mask_value=jnp.nan)._unscaled_cost_matrix
      return 1.0 / jnp.nanmean(geom)
    if self._scale_cost == "median":
      if not self.is_online:
        geom = self._masked_geom(mask_value=jnp.nan)
        return 1.0 / jnp.nanmedian(geom._unscaled_cost_matrix)
      raise NotImplementedError(
          "Using the median as scaling factor for "
          "the cost matrix with the online mode is not implemented."
      )
    if not hasattr(self.cost_fn, "norm"):
      raise ValueError("Cost function has no norm method.")
    norm_x = self.cost_fn.norm(self.x)
    norm_y = self.cost_fn.norm(self.y)
    if self._scale_cost == "max_norm":
      return 1.0 / jnp.maximum(norm_x.max(), norm_y.max())
    if self._scale_cost == "max_bound":
      if self.is_squared_euclidean:
        x_argmax = jnp.argmax(norm_x)
        y_argmax = jnp.argmax(norm_y)
        max_bound = (
            norm_x[x_argmax] + norm_y[y_argmax] +
            2 * jnp.sqrt(norm_x[x_argmax] * norm_y[y_argmax])
        )
        return 1.0 / max_bound
      raise NotImplementedError(
          "Using max_bound as scaling factor for "
          "the cost matrix when the cost is not squared euclidean "
          "is not implemented."
      )
    raise ValueError(f"Scaling {self._scale_cost} not implemented.")

  @property
  def kernel_matrix(self) -> Optional[jnp.ndarray]:  # noqa: D102
    return jnp.exp(-self.cost_matrix / self.epsilon)

  @property
  def shape(self) -> Tuple[int, int]:  # noqa: D102
    return self.x.shape[0], self.y.shape[0]

  @property
  def dtype(self) -> jnp.dtype:  # noqa: D102
    return self.x.dtype

  @property
  def is_symmetric(self) -> bool:  # noqa: D102
    return self.y is None or (
        jnp.all(self.x.shape == self.y.shape) and jnp.all(self.x == self.y)
    )

  @property
  def is_squared_euclidean(self) -> bool:  # noqa: D102
    return isinstance(self.cost_fn, costs.SqEuclidean)

  @property
  def can_LRC(self):  # noqa: D102
    return self.is_squared_euclidean and self._check_LRC_dim

  @property
  def _check_LRC_dim(self):
    (n, m), d = self.shape, self.x.shape[1]
    return n * m > (n + m) * d

  @property
  def cost_rank(self) -> int:  # noqa: D102
    return self.x.shape[1]

  @property
  def batch_size(self) -> Optional[int]:
    """Batch size for online mode."""
    if self._batch_size is None:
      return None
    n, m = self.shape
    return min(n, m, self._batch_size)

  @property
  def is_online(self) -> bool:
    """Whether the cost/kernel is computed on-the-fly."""
    return self.batch_size is not None



# Copyright OTT-JAX
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import abc
import functools
from typing import Any, Callable, Optional, Tuple, Union

import lineax as lx

import jax
import jax.numpy as jnp
import jax.tree_util as jtu

__all__ = [
    "PostComposition",
    "Regularization",
    "Orthogonal",
    "Quadratic",
    "L1",
    "SqL2",
    "STVS",
    "SqKOverlap",
]


class ProximalOperator(abc.ABC):
  """Proximal operator base class."""

  @abc.abstractmethod
  def __call__(self, x: jnp.ndarray) -> float:
    """Function.

    Args:
      x: Array of shape ``[d,]``.

    Returns:
      The value.
    """

  @abc.abstractmethod
  def prox(self, v: jnp.ndarray, tau: float = 1.0) -> jnp.ndarray:
    """Proximal operator.

    Args:
      v: Array of shape ``[d,]``.
      tau: Positive weight.

    Returns:
        The prox of ``v``.
    """

  def prox_dual(self, v: jnp.ndarray, tau: float = 1.0) -> jnp.ndarray:
    r"""Proximal operator of the convex conjugate.

    Uses Moreau's decomposition:

    .. math::
        v = \prox_{\tau f} \left(v\right) +
        \tau \prox_{\frac{1}{\tau} f^*} \left(\frac{v}{\tau}\right)

    Args:
      v: Array of shape ``[d,]``.
      tau: Positive weight.

    Returns:
      The prox dual of ``v``.
    """
    return v - tau * self.prox(v / tau, 1.0 / tau)

  def moreau_envelope(self, x: jnp.ndarray, tau: float = 1.0) -> jnp.ndarray:
    r"""Moreau Envelope.

    Uses Remark 12.24 from :cite:`bauschke:17`:

    .. math::
      {^\tau}f\left(x\right) = f\left(\prox_{\tau f}\left(x\right)\right)
      + \frac{1}{2\tau}\|x - \prox_{\tau f}\left(x\right)|_2^2

    Args:
      x: Array of shape ``[d,]``.
      tau: Positive weight.

    Returns:
      The Moreau Envelope of ``x``.
    """
    prox_x = self.prox(x, tau)
    return self(prox_x) + (1.0 / (2.0 * tau)) * jnp.sum((x - prox_x) ** 2)

  def tree_flatten(self):  # noqa: D102
    return (), {}

  @classmethod
  def tree_unflatten(cls, aux_data, children):  # noqa: D102
    return cls(*children, **aux_data)


@jtu.register_pytree_node_class
class PostComposition(ProximalOperator):
  r"""Postcomposition operator :math:`\alpha f\left(x\right) + b`.

  Args:
    f: Function :math:`f`.
    alpha: Scaling factor.
    b: Offset.
  """

  def __init__(self, f: ProximalOperator, alpha: float = 1.0, b: float = 0.0):
    super().__init__()
    self.f = f
    self.alpha = alpha
    self.b = b

  def __call__(self, x: jnp.ndarray) -> float:  # noqa: D102
    return self.alpha * self.f(x) + self.b

  def prox(self, v: jnp.ndarray, tau: float = 1.0) -> jnp.ndarray:  # noqa: D102
    return self.f.prox(v, tau * self.alpha)

  def tree_flatten(self):  # noqa: D102
    return (self.f, self.alpha, self.b), {}


@jtu.register_pytree_node_class
class Regularization(ProximalOperator):
  r"""Regularization operator :math:`f\left(x\right) + \frac{\rho}{2}\|x - a\|_2^2`.

  Args:
    f: Function :math:`f`.
    a: Offset :math:`a`. If :obj:`None`, use array of 0s.
    rho: Scaling factor.
  """  # noqa: E501

  def __init__(
      self,
      f: ProximalOperator,
      a: Optional[jnp.ndarray] = None,
      rho: float = 1.0,
  ):
    super().__init__()
    self.f = f
    self.a = a
    self.rho = rho

  def __call__(self, x: jnp.ndarray) -> float:  # noqa: D102
    norm = jnp.sum(x ** 2) if self.a is None else jnp.sum((x - self.a) ** 2)
    return self.f(x) + (0.5 * self.rho) * norm

  def prox(self, v: jnp.ndarray, tau: float = 1.0) -> jnp.ndarray:  # noqa: D102
    tau_tilde = tau / (1.0 + tau * self.rho)
    # (tau_tilde / tau) * v
    vv = 1.0 / (1 + tau * self.rho) * v
    if self.a is not None:
      vv = vv + (self.rho * tau_tilde) * self.a
    # section 2.2 of :cite:`parikh:14`
    return self.f.prox(vv, tau_tilde)

  def tree_flatten(self):  # noqa: D102
    return (self.f, self.a, self.rho), {}


@jtu.register_pytree_node_class
class Orthogonal(ProximalOperator):
  r"""Orthogonal operator :math:`f\left( Ax \right) + b`.

  The computation of the :meth:`prox` uses the
  Proposition 11 of :cite:`combettes:07`.

  Args:
    f: Function :math:`f` applied to :math:`Ax`.
    A: Linear operator :math:`A`.
    b: Offset :math:`b`. If :obj:`None`, use array of 0s.
    nu: Value for which :math:`AA^T = \nu I` holds.
  """

  def __init__(
      self,
      f: ProximalOperator,
      A: Optional[Union[jnp.ndarray, lx.AbstractLinearOperator]],
      b: Optional[jnp.ndarray] = None,
      nu: float = 1.0,
  ):
    assert nu > 0.0, nu
    super().__init__()
    self.f = f
    # AA^T = alpha I
    self.A = lx.MatrixLinearOperator(A) if isinstance(A, jnp.ndarray) else A
    self.b = b
    self.nu = nu

  def __call__(self, x: jnp.ndarray) -> float:  # noqa: D102
    z = self.A.mv(x)
    if self.b is not None:
      z = z + self.b
    return self.f(z)

  def prox(self, v: jnp.ndarray, tau: float = 1.0) -> jnp.ndarray:  # noqa: D102
    w = self.A.mv(v)
    if self.b is None:
      tmp = self.f.prox(w, tau * self.nu)
    else:
      tmp = self.f.prox(w + self.b, tau * self.nu) - self.b
    return v - (1.0 / self.nu) * (self.A.T.mv(w - tmp))

  @property
  def is_fully_orthogonal(self) -> bool:
    r"""Whether :math:`\nu = 1`."""
    return self.nu == 1.0

  def tree_flatten(self):  # noqa: D102
    return (self.f, self.A, self.b), {"nu": self.nu}


@jtu.register_pytree_node_class
class Quadratic(ProximalOperator):
  r"""Quadratic operator :math:`\frac{1}{2} \left<x, Q x\right> + b`.

  The matrix :math:`Q` is defined as:

  - :math:`Q := A` if not factored and not an orthogonal complement.
  - :math:`Q := A^{\perp}` if not factored and a complement.
  - :math:`Q := A^TA` if factored and not a complement.
  - :math:`Q := \left(A^{\perp}\right)^TA^{\perp}` if factored and
    a complement.

  Args:
    A: Linear operator :math:`A`. If :obj:`None`, use identity.
    b: Offset :math:`b`. If :obj:`None`, use array of 0s.
    is_complement: Whether to regularize in the orthogonal complement of
      :math:`A`, defined as :math:`A^{\perp} := I - A^T (AA^T)^{-1} A`.
    is_orthogonal: Whether :math:`AA^T = I`.
    is_factor: Whether to factor the matrix :math:`Q` as mentioned above.
    solver: Linear solver. If :obj:`None`, use :func:`lineax.linear_solve`.
  """

  def __init__(
      self,
      A: Optional[Union[jnp.ndarray, lx.AbstractLinearOperator]] = None,
      b: Optional[jnp.ndarray] = None,
      *,
      is_complement: bool = False,
      is_orthogonal: bool = False,
      is_factor: bool = False,
      solver: Optional[Callable[[lx.AbstractLinearOperator, jnp.ndarray],
                                jnp.ndarray]] = None,
  ):
    super().__init__()
    self.A = lx.MatrixLinearOperator(A) if isinstance(A, jnp.ndarray) else A
    self.b = b
    self._is_complement = is_complement
    self._is_orthogonal = is_orthogonal
    self._is_factor = is_factor
    self._solver = solver

  def __call__(self, x: jnp.ndarray) -> float:  # noqa: D102
    Q = self.Q
    y = 0.5 * (jnp.dot(x, x) if Q is None else jnp.dot(x, Q.mv(x)))
    return y if self.b is None else (y + jnp.dot(x, self.b))

  def prox(self, v: jnp.ndarray, tau: float = 1.0) -> jnp.ndarray:  # noqa: D102
    # section 6.1.1 in :cite:`parikh:14`
    Q = self.Q
    b = v if self.b is None else (v - tau * self.b)

    if Q is None:
      return (1.0 / (1.0 + tau)) * b

    iden = lx.IdentityLinearOperator(Q.out_structure())
    if self.is_factor:  # use matrix inversion lemma
      if self.is_complement:
        # eq. 14 in :cite:`klein:24`
        # A_comp = I - A^T(AA^T)^{-1}A
        # prox(v) = (I + tau A_comp^T A_comp)^{-1} (v - tau * b)
        op = iden + tau * (iden - self.A_comp)
        return (1.0 / (1.0 + tau)) * op.mv(b)
      if self.is_orthogonal:
        # https://en.wikipedia.org/wiki/Woodbury_matrix_identity
        op = iden - (tau / (1.0 + tau)) * (self.A.T @ self.A)
        return op.mv(b)

    A = iden + tau * Q

    if self._solver is None:
      # use default solver
      return lx.linear_solve(A, b).value
    return self._solver(A, b)

  @property
  def A_comp(self) -> Optional[lx.AbstractLinearOperator]:
    r"""Orthogonal complement :math:`A^{\perp}` of :math:`A`."""
    return _complement(
        self.A, self.is_orthogonal
    ) if self.is_complement else None

  @property
  def is_complement(self) -> bool:
    r"""Whether :attr:`Q` is defined using :math:`A_{\perp}` or :math:`A`."""
    return self.A is not None and self._is_complement

  @property
  def is_factor(self) -> bool:
    r"""Whether :attr:`Q` is factored."""
    return self.A is not None and self._is_factor

  @property
  def is_orthogonal(self) -> bool:
    r"""Whether :attr:`AA^T = I`."""
    return self.A is not None and self._is_orthogonal

  @property
  def Q(self) -> Optional[lx.AbstractLinearOperator]:
    r"""Linear operator :math:`Q`."""
    Q = self.A_comp if self.is_complement else self.A
    if Q is None:
      return None
    return (Q.T @ Q) if self.is_factor else Q

  def tree_flatten(self):  # noqa: D102
    return (self.A, self.b), {
        "is_complement": self.is_complement,
        "is_orthogonal": self.is_orthogonal,
        "is_factor": self.is_factor,
        "solver": self._solver
    }


@jtu.register_pytree_node_class
class L1(ProximalOperator):
  r"""L1-norm regularizer :math:`\ell_1`."""

  def __call__(self, x: jnp.ndarray) -> float:  # noqa: D102
    return jnp.linalg.norm(x, ord=1)

  def prox(self, v: jnp.ndarray, tau: float = 1.0) -> jnp.ndarray:  # noqa: D102
    return jnp.sign(v) * jax.nn.relu(jnp.abs(v) - tau)


@jtu.register_pytree_node_class
class SqL2(ProximalOperator):
  r"""Squared L2-norm regularizer :math:`\ell_2^2`.

  Args:
    A: Linear operator :math:`A` in :math:`\frac{1}{2} \left<x, A^TAx\right>`.
      If :obj:`None`, use identity.
    kwargs: Keyword arguments for :class:`Quadratic`.
  """

  def __init__(
      self,
      A: Optional[Union[jnp.ndarray, lx.AbstractLinearOperator]] = None,
      **kwargs: Any,
  ):
    super().__init__()
    self.f = Quadratic(A, is_factor=True, **kwargs)
    self._init_kwargs = kwargs

  def __call__(self, x: jnp.ndarray) -> float:  # noqa: D102
    return self.f(x)

  def prox(self, v: jnp.ndarray, tau: float = 1.0) -> jnp.ndarray:  # noqa: D102
    return self.f.prox(v, tau)

  def tree_flatten(self):  # noqa: D102
    return (self.f.A,), self._init_kwargs


@jtu.register_pytree_node_class
class STVS(ProximalOperator):
  r"""Soft thresholding operator with vanishing shrinkage regularizer :cite:`schreck:15`.

  The operator is defined as:

  .. math::
    \gamma^2 \mathbf{1}_d^T \left(\sigma(x) -
    \frac{1}{2} \exp\left(-2\sigma(x)\right) + \frac{1}{2}\right)

  where :math:`\sigma(x) := \text{asinh}\left(\frac{x}{2\gamma}\right)`.

  Args:
    gamma: Strength of the regularization.
  """  # noqa: E501

  def __init__(self, gamma: float = 1.0):
    super().__init__()
    self.gamma = gamma

  def __call__(self, x: jnp.ndarray) -> float:  # noqa: D102
    # Lemma 2.1 of `schreck:15`
    u = jnp.arcsinh(jnp.abs(x) / (2.0 * self.gamma))
    y = u - 0.5 * jnp.exp(-2.0 * u)
    return self.gamma ** 2 * jnp.sum(y + 0.5)  # make positive

  def prox(self, v: jnp.ndarray, tau: float = 1.0) -> jnp.ndarray:  # noqa: D102
    s = (tau * self.gamma) ** 2
    return jnp.where(v ** 2 <= s, 0.0, v - s / jnp.where(v == 0.0, 1.0, v))

  def tree_flatten(self):  # noqa: D102
    return (self.gamma,), {}


@jtu.register_pytree_node_class
class SqKOverlap(ProximalOperator):
  r"""Squared k-overlap norm regularizer :cite:`argyriou:12`.

  The regularizer is defined as:

  .. math::
    \frac{1}{2} \left(\|x\|_k^{\text{ov}}\right)^2

  where :math:`\left(\|x\|_k^{\text{ov}}\right)^2` is the squared k-overlap
  norm, defined in :cite:`argyriou:12`, def. 2.1.

  Args:
    k: Number of groups in :math:`[0, d)`, where :math:`d` is the dimensionality
      of the data.
  """

  def __init__(self, k: int):
    super().__init__()
    self.k = k

  def __call__(self, z: jnp.ndarray) -> float:  # noqa: D102
    # Prop 2.1 in :cite:`argyriou:12`
    k = self.k
    top_w = jax.lax.top_k(jnp.abs(z), k)[0]  # Fetch largest k values
    top_w = jnp.flip(top_w)  # Sort k-largest from smallest to largest
    # sum (dim - k) smallest values
    sum_bottom = jnp.sum(jnp.abs(z)) - jnp.sum(top_w)
    cumsum_top = jnp.cumsum(top_w)
    # Cesaro mean of top_w (each term offset with sum_bottom).
    cesaro = sum_bottom + cumsum_top
    cesaro /= jnp.arange(k) + 1
    # Choose first index satisfying constraint in Prop 2.1
    lower_bound = cesaro - top_w >= 0
    # Last upper bound is always True.
    upper_bound = jnp.concatenate(((top_w[1:] - cesaro[:-1]
                                    > 0), jnp.array((True,))))
    r = jnp.argmax(lower_bound * upper_bound)
    s = jnp.sum(jnp.where(jnp.arange(k) < k - r - 1, jnp.flip(top_w) ** 2, 0))

    return 0.5 * (s + (r + 1) * cesaro[r] ** 2)

  def prox(self, v: jnp.ndarray, tau: float = 1.0) -> float:  # noqa: D102

    @functools.partial(jax.vmap, in_axes=[0, None, None])
    def find_indices(r: int, l: jnp.ndarray,
                     z: jnp.ndarray) -> Tuple[jnp.ndarray, jnp.ndarray]:

      @functools.partial(jax.vmap, in_axes=[None, 0, None])
      def inner(r: int, l: int,
                z: jnp.ndarray) -> Tuple[jnp.ndarray, jnp.ndarray]:
        i = k - r - 1
        res = jnp.sum(z * ((i <= ixs) & (ixs < l)))
        res /= l - k + (beta + 1) * r + beta + 1

        cond1_left = jnp.logical_or(i == 0, (z[i - 1] / beta + 1) > res)
        cond1_right = res >= (z[i] / (beta + 1))
        cond1 = jnp.logical_and(cond1_left, cond1_right)

        cond2_left = z[l - 1] > res
        cond2_right = jnp.logical_or(l == d, res >= z[l])
        cond2 = jnp.logical_and(cond2_left, cond2_right)

        return res, cond1 & cond2

      return inner(r, l, z)

    # Alg. 1 of :cite:`argyriou:12`
    k, d, beta = self.k, v.shape[-1], 1.0 / tau

    ixs = jnp.arange(d)
    v, sgn = jnp.abs(v), jnp.sign(v)
    z_ixs = jnp.argsort(v)[::-1]
    z_sorted = v[z_ixs]

    # (k, d - k + 1)
    T, mask = find_indices(jnp.arange(k), jnp.arange(k, d + 1), z_sorted)
    (r,), (l,) = jnp.where(mask, size=1)  # size=1 for jitting
    T = T[r, l]

    q1 = (beta / (beta + 1)) * z_sorted * (ixs < (k - r - 1))
    q2 = (z_sorted - T) * jnp.logical_and((k - r - 1) <= ixs, ixs < (l + k))
    q = q1 + q2

    # change sign and reorder
    return sgn * q[jnp.argsort(z_ixs.astype(float))]

  def tree_flatten(self):  # noqa: D102
    return (), {"k": self.k}


def _invert(A: lx.AbstractLinearOperator) -> lx.MatrixLinearOperator:
  d = A.out_size()
  b = jnp.zeros(d)

  solve_fn = jax.vmap(lambda ix: lx.linear_solve(A, b.at[ix].set(1.0)).value)
  inv = solve_fn(jnp.arange(d))
  return lx.MatrixLinearOperator(inv)


@functools.partial(jax.jit, static_argnums=1)
def _complement(
    A: lx.AbstractLinearOperator, is_orthogonal: bool
) -> lx.AbstractLinearOperator:
  iden = lx.IdentityLinearOperator(A.in_structure())
  if is_orthogonal:
    # AA^T = I
    return iden - (A.T @ A)

  A_inv = _invert(lx.TaggedLinearOperator(A @ A.T, tags={lx.symmetric_tag}))
  return iden - A.T @ (A_inv @ A)



# Copyright OTT-JAX
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import functools
from typing import Any, Dict, Optional, Sequence, Tuple

import jax
import jax.numpy as jnp

import optax
from flax import linen as nn
from flax.core import frozen_dict
from flax.training import train_state

from ott import utils
from ott.geometry import geometry
from ott.initializers.linear import initializers
from ott.problems.linear import linear_problem
from ott.solvers.linear import sinkhorn

__all__ = ["MetaInitializer"]


@jax.tree_util.register_pytree_node_class
class MetaInitializer(initializers.DefaultInitializer):
  """Meta OT Initializer with a fixed geometry :cite:`amos:22`.

  This initializer consists of a predictive model that outputs the
  :math:`f` duals to solve the entropy-regularized OT problem given
  input probability weights ``a`` and ``b``, and a given (assumed to be
  fixed) geometry ``geom``.

  The model's parameters are learned using a training set of OT
  instances (multiple pairs of probability weights), that assume the
  **same** geometry ``geom`` is used throughout, both for training and
  evaluation.

  Args:
    geom: The fixed geometry of the problem instances.
    meta_model: The model to predict the potential :math:`f` from the measures.
    opt: The optimizer to update the parameters. If :obj:`None`, use
      :func:`optax.adam` with :math:`0.001` learning rate.
    rng: The PRNG key to use for initializing the model.
    state: The training state of the model to start from.

  Examples:
    The following code shows a simple
    example of using ``update`` to train the model, where
    ``a`` and ``b`` are the weights of the measures and
    ``geom`` is the fixed geometry.

    .. code-block:: python

      meta_initializer = init_lib.MetaInitializer(geom)
      while training():
        a, b = sample_batch()
        loss, init_f, meta_initializer.state = meta_initializer.update(
          meta_initializer.state, a=a, b=b
        )
  """

  def __init__(
      self,
      geom: geometry.Geometry,
      meta_model: nn.Module,
      opt: Optional[optax.GradientTransformation
                   ] = optax.adam(learning_rate=1e-3),  # noqa: B008
      rng: Optional[jax.Array] = None,
      state: Optional[train_state.TrainState] = None,
  ):
    self.geom = geom
    self.opt = opt
    self.rng = utils.default_prng_key(rng)

    na, nb = geom.shape
    self.meta_model = meta_model

    if state is None:
      # Initialize the model's training state.
      placeholder = jnp.concatenate([jnp.zeros(na), jnp.zeros(nb)], axis=-1)
      params = self.meta_model.init(self.rng, placeholder)["params"]
      self.state = train_state.TrainState.create(
          apply_fn=self.meta_model.apply, params=params, tx=opt
      )
    else:
      self.state = state

    self.update_impl = self._get_update_fn()

  def update(
      self, state: train_state.TrainState, a: jnp.ndarray, b: jnp.ndarray
  ) -> Tuple[jnp.ndarray, jnp.ndarray, train_state.TrainState]:
    r"""Update the meta model with the dual objective.

    The goal is for the model to match the optimal duals, i.e.,
    :math:`\hat f_\theta \approx f^\star`.
    This can be done by training the predictions of :math:`\hat f_\theta`
    to optimize the dual objective, which :math:`f^\star` also optimizes for.
    The overall learning setup can thus be written as:

    .. math::
      \min_\theta\; {\mathbb E}_{(\alpha,\beta)\sim{\mathcal{D}}}\;
      J(\hat f_\theta(a, b); \alpha, \beta),

    where :math:`a,b` are the probabilities of the measures :math:`\alpha,\beta`
    ,:math:`\mathcal{D}` is a meta distribution of optimal transport problems,

    .. math::
      -J(f; \alpha, \beta, c) := \langle f, a\rangle + \langle g, b \rangle -
      \varepsilon\left\langle \exp\{f/\varepsilon\}, K\exp\{g/\varepsilon\}
      \right\rangle

    is the entropic dual objective,
    and :math:`K_{i,j} := -C_{i,j}/\varepsilon` is the *Gibbs kernel*.

    Args:
      state: Optimizer state of the meta model.
      a: Probabilities of the :math:`\alpha` measure's atoms.
      b: Probabilities of the :math:`\beta` measure's atoms.

    Returns:
      The training loss, :math:`f`, and updated state.
    """
    return self.update_impl(state, a, b)

  def init_dual_a(  # noqa: D102
      self,
      ot_prob: linear_problem.LinearProblem,
      lse_mode: bool,
      rng: Optional[jax.Array] = None,
  ) -> jnp.ndarray:
    del rng
    # Detect if the problem is batched.
    assert ot_prob.a.ndim in (1, 2)
    assert ot_prob.b.ndim in (1, 2)
    vmap_a_val = 0 if ot_prob.a.ndim == 2 else None
    vmap_b_val = 0 if ot_prob.b.ndim == 2 else None

    if vmap_a_val is not None or vmap_b_val is not None:
      compute_f_maybe_batch = jax.vmap(
          self._compute_f, in_axes=(vmap_a_val, vmap_b_val, None)
      )
    else:
      compute_f_maybe_batch = self._compute_f

    init_f = compute_f_maybe_batch(ot_prob.a, ot_prob.b, self.state.params)
    return init_f if lse_mode else ot_prob.geom.scaling_from_potential(init_f)

  def _get_update_fn(self):
    """Return the implementation (and jitted) update function."""

    def dual_obj_loss_single(params, a, b):
      f_pred = self._compute_f(a, b, params)
      g_pred = self.geom.update_potential(
          f_pred, jnp.zeros_like(b), jnp.log(b), 0, axis=0
      )
      g_pred = jnp.where(jnp.isfinite(g_pred), g_pred, 0.0)

      ot_prob = linear_problem.LinearProblem(geom=self.geom, a=a, b=b)
      dual_obj = sinkhorn.compute_kl_reg_cost(
          f_pred, g_pred, ot_prob, lse_mode=True
      )
      loss = -dual_obj
      return loss, f_pred

    def loss_batch(params, a, b):
      loss_fn = functools.partial(dual_obj_loss_single, params=params)
      loss, f_pred = jax.vmap(loss_fn)(a=a, b=b)
      return jnp.mean(loss), f_pred

    @jax.jit
    def update(state, a, b):
      a = jnp.atleast_2d(a)
      b = jnp.atleast_2d(b)
      grad_fn = jax.value_and_grad(loss_batch, has_aux=True)
      (loss, init_f), grads = grad_fn(state.params, a, b)
      return loss, init_f, state.apply_gradients(grads=grads)

    return update

  def _compute_f(
      self, a: jnp.ndarray, b: jnp.ndarray,
      params: frozen_dict.FrozenDict[str, jnp.ndarray]
  ) -> jnp.ndarray:
    r"""Predict the optimal :math:`f` potential.

    Args:
      a: Probabilities of the :math:`\alpha` measure's atoms.
      b: Probabilities of the :math:`\beta` measure's atoms.
      params: The parameters of the Meta model.

    Returns:
      The :math:`f` potential.
    """
    return self.meta_model.apply({"params": params},
                                 jnp.concatenate([a, b], axis=-1))

  def tree_flatten(self) -> Tuple[Sequence[Any], Dict[str, Any]]:  # noqa: D102
    return [self.geom, self.meta_model, self.opt], {
        "rng": self.rng,
        "state": self.state
    }



# Copyright OTT-JAX
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any, Callable, Optional, Tuple

import jax
import jax.numpy as jnp

from ott import utils
from ott.geometry import costs, pointcloud
from ott.solvers import linear
from ott.solvers.linear import univariate

__all__ = ["random_proj_sphere", "sliced_wasserstein"]

Projector = Callable[[jnp.ndarray, int, jax.Array], jnp.ndarray]


def random_proj_sphere(
    x: jnp.ndarray,
    n_proj: int = 1000,
    rng: Optional[jax.Array] = None
) -> jnp.ndarray:
  """Project data on directions sampled randomly from sphere.

  Args:
    x: Array of size ``[n, dim]``.
    n_proj: Number of randomly generated projections.
    rng: Key used to sample feature extractors.

  Returns:
    Array of size ``[n, n_proj]`` features.
  """
  rng = utils.default_prng_key(rng)
  dim = x.shape[-1]
  proj_m = jax.random.normal(rng, (n_proj, dim))
  proj_m /= jnp.linalg.norm(proj_m, axis=1, keepdims=True)
  return x @ proj_m.T


def sliced_wasserstein(
    x: jnp.ndarray,
    y: jnp.ndarray,
    a: Optional[jnp.ndarray] = None,
    b: Optional[jnp.ndarray] = None,
    cost_fn: Optional[costs.CostFn] = None,
    proj_fn: Optional[Projector] = None,
    weights: Optional[jnp.ndarray] = None,
    return_transport: bool = False,
    return_dual_variables: bool = False,
    **kwargs: Any,
) -> Tuple[jnp.ndarray, univariate.UnivariateOutput]:
  r"""Compute the Sliced Wasserstein distance between two weighted point clouds.

  Follows the approach outlined in :cite:`rabin:12` to compute a proxy for OT
  distances that relies on creating features (possibly randomly) for data,
  through e.g., projections, and then sum the 1D Wasserstein distances between
  these features' univariate distributions on both source and target samples.

  Args:
    x: Array of shape ``[n, dim]`` of source points' coordinates.
    y: Array of shape ``[m, dim]`` of target points' coordinates.
    a: Array of shape ``[n,]`` of source probability weights.
    b: Array of shape ``[m,]`` of target probability weights.
    cost_fn: Cost function. Must be a submodular function of two real arguments,
      i.e. such that :math:`\partial c(x,y)/\partial x \partial y <0`. If
      :obj:`None`, use :class:`~ott.geometry.costs.SqEuclidean`.
    proj_fn: Projection function, mapping any ``[b, dim]`` matrix of coordinates
      to ``[b, n_proj]`` matrix of features, on which 1D transports (for
      ``n_proj`` directions) are subsequently computed independently.
      By default, use :func:`~ott.tools.sliced.random_proj_sphere`.
    weights: Array of shape ``[n_proj,]`` of weights used to average the
      ``n_proj`` 1D Wasserstein contributions (one for each feature) and form
      the sliced Wasserstein distance. Uniform by default, resulting in average
      of all these values.
    return_transport: Whether to store ``n_proj`` transport plans in the output.
    return_dual_variables: Whether to store ``n_proj`` pairs of dual vectors
      in the output.
    kwargs: Keyword arguments to ``proj_fn``. Could for instance
      include, as done with default projector, number of ``n_proj`` projections,
      as well as a ``rng`` key to sample as many directions.

  Returns:
    The sliced Wasserstein distance with the corresponding output object.
  """
  if proj_fn is None:
    proj_fn = random_proj_sphere

  x_proj, y_proj = proj_fn(x, **kwargs), proj_fn(y, **kwargs),
  geom = pointcloud.PointCloud(x_proj, y_proj, cost_fn=cost_fn)

  out = linear.solve_univariate(
      geom,
      a,
      b,
      return_transport=return_transport,
      return_dual_variables=return_dual_variables
  )
  return jnp.average(out.ot_costs, weights=weights), out



# Copyright OTT-JAX
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from . import (
    acceleration,
    continuous_barycenter,
    discrete_barycenter,
    implicit_differentiation,
    lr_utils,
    sinkhorn,
    sinkhorn_lr,
    univariate,
)
from ._solve import solve, solve_univariate

__all__ = [
    "acceleration",
    "continuous_barycenter",
    "discrete_barycenter",
    "implicit_differentiation",
    "lr_utils",
    "sinkhorn",
    "sinkhorn_lr",
    "univariate",
    "solve",
    "solve_univariate",
]



# Copyright OTT-JAX
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Callable, Optional

import jax.numpy as jnp
import jax.tree_util as jtu

from ott.geometry import costs, pointcloud
from ott.problems.linear import linear_problem
from ott.solvers.linear import univariate

__all__ = ["UnivariateWasserstein"]


@jtu.register_pytree_node_class
class UnivariateWasserstein(costs.CostFn):
  """1D Wasserstein cost for two 1D distributions.

  This ground cost between considers vectors as a family of values.
  The Wasserstein distance between them is the 1D OT cost, using a user-defined
  ground cost.

  Args:
    solve_fn: 1D optimal transport solver, e.g.,
      :func:`~ott.solvers.linear.univariate.uniform_distance`.
    ground_cost: Cost used to compute the 1D optimal transport between vectors.
      Should be a translation-invariant (TI) cost for correctness.
      If :obj:`None`, defaults to :class:`~ott.geometry.costs.SqEuclidean`.
  """

  def __init__(
      self,
      solve_fn: Callable[[linear_problem.LinearProblem],
                         univariate.UnivariateOutput],
      ground_cost: Optional[costs.TICost] = None,
  ):
    super().__init__()
    self.ground_cost = (
        costs.SqEuclidean() if ground_cost is None else ground_cost
    )
    self._solve_fn = solve_fn

  def __call__(self, x: jnp.ndarray, y: jnp.ndarray) -> float:
    """Wasserstein distance between :math:`x` and :math:`y` seen as a 1D dist.

    Args:
      x: Array of shape ``[n,]``.
      y: Array of shape ``[m,]``.

    Returns:
      The transport cost.
    """
    geom = pointcloud.PointCloud(
        x[:, None], y[:, None], cost_fn=self.ground_cost
    )
    prob = linear_problem.LinearProblem(geom)
    out = self._solve_fn(prob)
    return jnp.squeeze(out.ot_costs)

  def tree_flatten(self):  # noqa: D102
    return (self.ground_cost,), (self._solve_fn,)

  @classmethod
  def tree_unflatten(cls, aux_data, children):  # noqa: D102
    return cls(solve_fn=aux_data[0], ground_cost=children[0])



# Copyright OTT-JAX
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import (
    Any,
    Callable,
    Literal,
    Mapping,
    NamedTuple,
    Optional,
    Sequence,
    Tuple,
    Union,
)

import jax
import jax.numpy as jnp
import jax.scipy as jsp
import numpy as np

from ott import utils
from ott.geometry import geometry
from ott.initializers.linear import initializers as init_lib
from ott.math import fixed_point_loop
from ott.math import unbalanced_functions as uf
from ott.math import utils as mu
from ott.problems.linear import linear_problem, potentials
from ott.solvers.linear import acceleration
from ott.solvers.linear import implicit_differentiation as implicit_lib

__all__ = ["Sinkhorn", "SinkhornOutput"]

ProgressCallbackFn_t = Callable[
    [Tuple[np.ndarray, np.ndarray, np.ndarray, "SinkhornState"]], None]


class SinkhornState(NamedTuple):
  """Holds the state variables used to solve OT with Sinkhorn."""

  potentials: Tuple[jnp.ndarray, ...]
  errors: Optional[jnp.ndarray] = None
  old_fus: Optional[jnp.ndarray] = None
  old_mapped_fus: Optional[jnp.ndarray] = None

  def set(self, **kwargs: Any) -> "SinkhornState":
    """Return a copy of self, with potential overwrites."""
    return self._replace(**kwargs)

  def solution_error(
      self,
      ot_prob: linear_problem.LinearProblem,
      norm_error: Sequence[int],
      *,
      lse_mode: bool,
      parallel_dual_updates: bool,
      recenter: bool,
  ) -> jnp.ndarray:
    """State dependent function to return error."""
    fu, gv = self.fu, self.gv
    if recenter and lse_mode:
      fu, gv = self.recenter(fu, gv, ot_prob=ot_prob)

    return solution_error(
        fu,
        gv,
        ot_prob,
        norm_error=norm_error,
        lse_mode=lse_mode,
        parallel_dual_updates=parallel_dual_updates
    )

  def compute_kl_reg_cost(  # noqa: D102
      self, ot_prob: linear_problem.LinearProblem, lse_mode: bool
  ) -> jnp.ndarray:
    return compute_kl_reg_cost(self.fu, self.gv, ot_prob, lse_mode)

  def recenter(
      self,
      f: jnp.ndarray,
      g: jnp.ndarray,
      ot_prob: linear_problem.LinearProblem,
  ) -> Tuple[jnp.ndarray, jnp.ndarray]:
    """Re-center dual potentials.

    If the ``ot_prob`` is balanced, the ``f`` potential is zero-centered.
    Otherwise, use Prop. 2 of :cite:`sejourne:22` re-center the potentials iff
    ``tau_a < 1`` and ``tau_b < 1``.

    Args:
      f: The first dual potential.
      g: The second dual potential.
      ot_prob: Linear OT problem.

    Returns:
      The centered potentials.
    """
    if ot_prob.is_balanced:
      # center the potentials for numerical stability
      is_finite = jnp.isfinite(f)
      shift = jnp.sum(jnp.where(is_finite, f, 0.0)) / jnp.sum(is_finite)
      return f - shift, g + shift

    if ot_prob.tau_a == 1.0 or ot_prob.tau_b == 1.0:
      # re-centering wasn't done during the lse-step, ignore
      return f, g

    rho_a = uf.rho(ot_prob.epsilon, ot_prob.tau_a)
    rho_b = uf.rho(ot_prob.epsilon, ot_prob.tau_b)
    tau = rho_a * rho_b / (rho_a + rho_b)

    shift = tau * (
        mu.logsumexp(-f / rho_a, b=ot_prob.a) -
        mu.logsumexp(-g / rho_b, b=ot_prob.b)
    )
    return f + shift, g - shift

  @property
  def fu(self) -> jnp.ndarray:
    """The first dual potential or scaling."""
    return self.potentials[0]

  @property
  def gv(self) -> jnp.ndarray:
    """The second dual potential or scaling."""
    return self.potentials[1]


def solution_error(
    f_u: jnp.ndarray,
    g_v: jnp.ndarray,
    ot_prob: linear_problem.LinearProblem,
    *,
    norm_error: Sequence[int],
    lse_mode: bool,
    parallel_dual_updates: bool,
) -> jnp.ndarray:
  """Given two potential/scaling solutions, computes deviation to optimality.

  When the ``ot_prob`` problem is balanced and the usual Sinkhorn updates are
  used, this is simply deviation of the coupling's marginal to ``ot_prob.b``.
  This is the case because the second (and last) update of the Sinkhorn
  algorithm equalizes the row marginal of the coupling to ``ot_prob.a``. To
  simplify the logic, this is parameterized by checking whether
  `parallel_dual_updates = False`.

  When that flag is `True`, or when the problem is unbalanced,
  additional quantities to qualify optimality must be taken into account.

  Args:
    f_u: jnp.ndarray, potential or scaling
    g_v: jnp.ndarray, potential or scaling
    ot_prob: linear OT problem
    norm_error: int, p-norm used to compute error.
    lse_mode: True if log-sum-exp operations, False if kernel vector products.
    parallel_dual_updates: Whether potentials/scalings were computed in
      parallel.

  Returns:
    a positive number quantifying how far from optimality current solution is.
  """
  if ot_prob.is_balanced and not parallel_dual_updates:
    return marginal_error(
        f_u, g_v, ot_prob.b, ot_prob.geom, 0, norm_error, lse_mode
    )

  # In the unbalanced case, we compute the norm of the gradient.
  # the gradient is equal to the marginal of the current plan minus
  # the gradient of < z, rho_z(exp^(-h/rho_z) -1> where z is either a or b
  # and h is either f or g. Note this is equal to z if rho_z → inf, which
  # is the case when tau_z → 1.0
  if lse_mode:
    grad_a = uf.grad_of_marginal_fit(
        ot_prob.a, f_u, ot_prob.tau_a, ot_prob.epsilon
    )
    grad_b = uf.grad_of_marginal_fit(
        ot_prob.b, g_v, ot_prob.tau_b, ot_prob.epsilon
    )
  else:
    u = ot_prob.geom.potential_from_scaling(f_u)
    v = ot_prob.geom.potential_from_scaling(g_v)
    grad_a = uf.grad_of_marginal_fit(
        ot_prob.a, u, ot_prob.tau_a, ot_prob.epsilon
    )
    grad_b = uf.grad_of_marginal_fit(
        ot_prob.b, v, ot_prob.tau_b, ot_prob.epsilon
    )
  err = marginal_error(f_u, g_v, grad_a, ot_prob.geom, 1, norm_error, lse_mode)
  err += marginal_error(f_u, g_v, grad_b, ot_prob.geom, 0, norm_error, lse_mode)
  return err


def marginal_error(
    f_u: jnp.ndarray,
    g_v: jnp.ndarray,
    target: jnp.ndarray,
    geom: geometry.Geometry,
    axis: int = 0,
    norm_error: Sequence[int] = (1,),
    lse_mode: bool = True
) -> jnp.asarray:
  """Output how far Sinkhorn solution is w.r.t target.

  Args:
    f_u: a vector of potentials or scalings for the first marginal.
    g_v: a vector of potentials or scalings for the second marginal.
    target: target marginal.
    geom: Geometry object.
    axis: axis (0 or 1) along which to compute marginal.
    norm_error: (tuple of int) p's to compute p-norm between marginal/target
    lse_mode: whether operating on scalings or potentials

  Returns:
    Array of floats, quantifying difference between target / marginal.
  """
  if lse_mode:
    marginal = geom.marginal_from_potentials(f_u, g_v, axis=axis)
  else:
    marginal = geom.marginal_from_scalings(f_u, g_v, axis=axis)
  norm_error = jnp.asarray(norm_error)
  return jnp.sum(
      jnp.abs(marginal - target) ** norm_error[:, jnp.newaxis], axis=1
  ) ** (1.0 / norm_error)


def compute_kl_reg_cost(
    f: jnp.ndarray, g: jnp.ndarray, ot_prob: linear_problem.LinearProblem,
    lse_mode: bool
) -> jnp.ndarray:
  r"""Compute objective of Sinkhorn for OT problem given dual solutions.

  The objective is evaluated for dual solution ``f`` and ``g``, using
  information contained in  ``ot_prob``. The objective is the regularized
  optimal transport cost (i.e. the cost itself plus entropic and unbalanced
  terms). Situations where marginals ``a`` or ``b`` in ``ot_prob`` have zero
  coordinates are reflected in minus infinity entries in their corresponding
  dual potentials. To avoid NaN that may result when multiplying 0's by infinity
  values, ``jnp.where`` is used to cancel these contributions.

  Args:
    f: jnp.ndarray, potential
    g: jnp.ndarray, potential
    ot_prob: linear optimal transport problem.
    lse_mode: bool, whether to compute total mass in lse or kernel mode.

  Returns:
    The regularized transport cost.
  """
  supp_a = ot_prob.a > 0
  supp_b = ot_prob.b > 0
  fa = ot_prob.geom.potential_from_scaling(ot_prob.a)
  if ot_prob.tau_a == 1.0:
    div_a = jnp.sum(jnp.where(supp_a, ot_prob.a * (f - fa), 0.0))
  else:
    rho_a = uf.rho(ot_prob.epsilon, ot_prob.tau_a)
    div_a = -jnp.sum(
        jnp.where(supp_a, ot_prob.a * uf.phi_star(-(f - fa), rho_a), 0.0)
    )

  gb = ot_prob.geom.potential_from_scaling(ot_prob.b)
  if ot_prob.tau_b == 1.0:
    div_b = jnp.sum(jnp.where(supp_b, ot_prob.b * (g - gb), 0.0))
  else:
    rho_b = uf.rho(ot_prob.epsilon, ot_prob.tau_b)
    div_b = -jnp.sum(
        jnp.where(supp_b, ot_prob.b * uf.phi_star(-(g - gb), rho_b), 0.0)
    )

  # Using https://arxiv.org/pdf/1910.12958v2.pdf (24)
  if lse_mode:
    total_sum = jnp.sum(ot_prob.geom.marginal_from_potentials(f, g))
  else:
    u = ot_prob.geom.scaling_from_potential(f)
    v = ot_prob.geom.scaling_from_potential(g)
    total_sum = jnp.sum(ot_prob.geom.marginal_from_scalings(u, v))

  return div_a + div_b + ot_prob.epsilon * (
      jnp.sum(ot_prob.a) * jnp.sum(ot_prob.b) - total_sum
  )


class SinkhornOutput(NamedTuple):
  """Holds the output of a Sinkhorn solver applied to a problem.

  Objects of this class contain both solutions and problem definition of a
  regularized OT problem, along several methods that can be used to access its
  content, to, for instance, materialize an OT matrix or apply it to a vector
  (without having to materialize it when not needed).

  Args:
    potentials: list of optimal dual variables, two vector of size
      ``ot.prob.shape[0]`` and ``ot.prob.shape[1]`` returned by Sinkhorn
    errors: vector or errors, along iterations. This vector is of size
      ``max_iterations // inner_iterations`` where those were the parameters
      passed on to the :class:`~ott.solvers.linear.sinkhorn.Sinkhorn` solver.
      For each entry indexed at ``i``, ``errors[i]`` can be either a real
      non-negative value (meaning the algorithm recorded that error at the
      ``i * inner_iterations`` iteration), a ``jnp.inf`` value (meaning the
      algorithm computed that iteration but did not compute its error, because,
      for instance, ``i < min_iterations // inner_iterations``), or a ``-1``,
      meaning that execution was terminated before that iteration, because the
      criterion was found to be smaller than ``threshold``.
    reg_ot_cost: the regularized optimal transport cost. By default this is
      the linear contribution + KL term. See
      :attr:`~ott.solvers.linear.sinkhorn.SinkhornOutput.ent_reg_cost`,
      :attr:`~ott.solvers.linear.sinkhorn.SinkhornOutput.primal_cost` and
      :attr:`~ott.solvers.linear.sinkhorn.SinkhornOutput.dual_cost` for other
      objective values.
    ot_prob: stores the definition of the OT problem, including geometry,
      marginals, unbalanced regularizers, etc.
    threshold: convergence threshold used to control the termination of the
      algorithm.
    converged: whether the output corresponds to a solution whose error is
      below the convergence threshold.
    inner_iterations: number of iterations that were run between two
      computations of errors.
  """

  potentials: Tuple[jnp.ndarray, ...]
  errors: Optional[jnp.ndarray] = None
  reg_ot_cost: Optional[jnp.ndarray] = None
  ot_prob: Optional[linear_problem.LinearProblem] = None
  threshold: Optional[jnp.ndarray] = None
  converged: Optional[bool] = None
  inner_iterations: Optional[int] = None

  def set(self, **kwargs: Any) -> "SinkhornOutput":
    """Return a copy of self, with potential overwrites."""
    return self._replace(**kwargs)

  def set_cost(  # noqa: D102
      self, ot_prob: linear_problem.LinearProblem, lse_mode: bool,
      use_danskin: bool
  ) -> "SinkhornOutput":
    f = jax.lax.stop_gradient(self.f) if use_danskin else self.f
    g = jax.lax.stop_gradient(self.g) if use_danskin else self.g
    return self.set(reg_ot_cost=compute_kl_reg_cost(f, g, ot_prob, lse_mode))

  @property
  def dual_cost(self) -> jnp.ndarray:
    """Return dual transport cost, without considering regularizer."""
    a, b = self.ot_prob.a, self.ot_prob.b
    dual_cost = jnp.sum(jnp.where(a > 0.0, a * self.f, 0))
    dual_cost += jnp.sum(jnp.where(b > 0.0, b * self.g, 0))
    return dual_cost

  @property
  def primal_cost(self) -> jnp.ndarray:
    """Return transport cost of current transport solution at geometry."""
    return self.transport_cost_at_geom(other_geom=self.geom)

  @property
  def ent_reg_cost(self) -> jnp.ndarray:
    r"""Entropy regularized cost.

    This outputs

    .. math::

      \langle P^{\star},C\rangle - \varepsilon H(P^{\star}) +
      \rho_a\text{KL}(P^{\star} 1|a) + \rho_b\text{KL}(1^T P^{\star}|b),

    where :math:`P^{\star}, a, b` is the coupling returned by the
    :class:`~ott.solvers.linear.sinkhorn.Sinkhorn` and the two marginal weight
    vectors; :math:`\rho_a=\varepsilon \tau_a / (1-\tau_a)` and
    :math:`\rho_b=\varepsilon \tau_b / (1-\tau_b)` are obtained when the problem
    is unbalanced from parameters ``tau_a`` and ``tau_b``. Note that the last
    two terms vanish in the balanced case, when ``tau_a==tau_b==1``.
    """
    ent_a = jnp.sum(jsp.special.entr(self.ot_prob.a))
    ent_b = jnp.sum(jsp.special.entr(self.ot_prob.b))
    return self.reg_ot_cost - self.geom.epsilon * (ent_a + ent_b)

  @property
  def kl_reg_cost(self) -> jnp.ndarray:
    r"""KL regularized OT transport cost.

    This outputs

    .. math::

      \langle P^{\star}, C \rangle + \varepsilon KL(P^{\star},ab^T) +
      \rho_a\text{KL}(P^{\star} 1|a) + \rho_b\text{KL}(1^T P^{\star}|b),

    where :math:`P^{\star}, a, b` are the coupling returned by the
    :class:`~ott.solvers.linear.sinkhorn.Sinkhorn` algorithm and the two
    marginal weight vectors, respectively, and
    :math:`\rho_a=\varepsilon \tau_a / (1-\tau_a)` and
    :math:`\rho_b=\varepsilon \tau_b / (1-\tau_b)` are obtained when the problem
    is unbalanced from parameters ``tau_a`` and ``tau_b``. Note that the last
    two terms vanish in the balanced case, when ``tau_a==tau_b==1``. This
    quantity coincides with :attr:`reg_ot_cost`, which is computed using
    dual variables.
    """
    return self.reg_ot_cost

  def transport_cost_at_geom(
      self, other_geom: geometry.Geometry
  ) -> jnp.ndarray:
    r"""Return bare transport cost of current solution at any geometry.

    In order to compute cost, we check first if the geometry can be converted
    to a low-rank cost geometry in order to speed up computations, without
    having to materialize the full cost matrix. If this is not possible,
    we resort to instantiating both transport matrix and cost matrix.

    Args:
      other_geom: geometry whose cost matrix is used to evaluate the transport
        cost.

    Returns:
      the transportation cost at :math:`C`, i.e. :math:`\langle P, C \rangle`.
    """
    # TODO(cuturi): handle online mode for non Euclidean pointcloud geometries.
    # TODO(michalk8): handle SqEucl point cloud is not converted to LRCGeom
    if other_geom.can_LRC:
      geom = other_geom.to_LRCGeometry()
      return jnp.sum(self.apply(geom.cost_1.T) * geom.cost_2.T)
    return jnp.sum(self.matrix * other_geom.cost_matrix)

  @property
  def geom(self) -> geometry.Geometry:  # noqa: D102
    return self.ot_prob.geom

  @property
  def a(self) -> jnp.ndarray:  # noqa: D102
    return self.ot_prob.a

  @property
  def b(self) -> jnp.ndarray:  # noqa: D102
    return self.ot_prob.b

  @property
  def n_iters(self) -> int:  # noqa: D102
    """Returns the total number of iterations that were needed to terminate."""
    return jnp.sum(self.errors != -1) * self.inner_iterations

  @property
  def scalings(self) -> Tuple[jnp.ndarray, jnp.ndarray]:  # noqa: D102
    u = self.ot_prob.geom.scaling_from_potential(self.f)
    v = self.ot_prob.geom.scaling_from_potential(self.g)
    return u, v

  @property
  def matrix(self) -> jnp.ndarray:
    """Transport matrix if it can be instantiated."""
    try:
      return self.ot_prob.geom.transport_from_potentials(self.f, self.g)
    except ValueError:
      return self.ot_prob.geom.transport_from_scalings(*self.scalings)

  @property
  def transport_mass(self) -> jnp.ndarray:
    """Sum of transport matrix."""
    return self.marginal(0).sum()

  def apply(
      self,
      inputs: jnp.ndarray,
      axis: int = 0,
      lse_mode: bool = True
  ) -> jnp.ndarray:
    """Apply the transport to a ndarray; axis=1 for its transpose."""
    geom = self.ot_prob.geom
    if lse_mode:
      return geom.apply_transport_from_potentials(
          self.f, self.g, inputs, axis=axis
      )
    u = geom.scaling_from_potential(self.f)
    v = geom.scaling_from_potential(self.g)
    return geom.apply_transport_from_scalings(u, v, inputs, axis=axis)

  def marginal(self, axis: int) -> jnp.ndarray:  # noqa: D102
    return self.ot_prob.geom.marginal_from_potentials(self.f, self.g, axis=axis)

  def cost_at_geom(self, other_geom: geometry.Geometry) -> jnp.ndarray:
    """Return reg-OT cost for matrix, evaluated at other cost matrix."""
    return (
        jnp.sum(self.matrix * other_geom.cost_matrix) -
        self.geom.epsilon * jnp.sum(jax.scipy.special.entr(self.matrix))
    )

  def to_dual_potentials(self) -> potentials.EntropicPotentials:
    """Return the entropic map estimator."""
    return potentials.EntropicPotentials(self.f, self.g, self.ot_prob)

  @property
  def f(self) -> jnp.ndarray:
    """The first dual potential."""
    return self.potentials[0]

  @property
  def g(self) -> jnp.ndarray:
    """The second dual potential."""
    return self.potentials[1]


@jax.tree_util.register_pytree_node_class
class Sinkhorn:
  r"""Sinkhorn solver.

  The Sinkhorn algorithm is a fixed point iteration that solves a regularized
  optimal transport (reg-OT) problem between two measures.
  The optimization variables are a pair of vectors (called potentials, or
  scalings when parameterized as exponential of the former). Calling this
  function returns therefore a pair of optimal vectors. In addition to these,
  it also returns the objective value achieved by these optimal vectors;
  a vector of size ``max_iterations/inner_iterations`` that records the vector
  of values recorded to monitor convergence, throughout the execution of the
  algorithm (padded with `-1` if convergence happens before), as well as a
  boolean to signify whether the algorithm has converged within the number of
  iterations specified by the user.

  The reg-OT problem is specified by two measures, of respective sizes ``n`` and
  ``m``. From the viewpoint of the ``sinkhorn`` function, these two measures are
  only seen through a triplet (``geom``, ``a``, ``b``), where ``geom`` is a
  ``Geometry`` object, and ``a`` and ``b`` are weight vectors of respective
  sizes ``n`` and ``m``. Starting from two initial values for those potentials
  or scalings (both can be defined by the user by passing value in
  ``init_dual_a`` or ``init_dual_b``), the Sinkhorn algorithm will use
  elementary operations that are carried out by the ``geom`` object.

  Math:
    Given a geometry ``geom``, which provides a cost matrix :math:`C` with its
    regularization parameter :math:`\varepsilon`, (or a kernel matrix :math:`K`)
    the reg-OT problem consists in finding two vectors `f`, `g` of size ``n``,
    ``m`` that maximize the following criterion.

    .. math::

      \arg\max_{f, g}{- \langle a, \phi_a^{*}(-f) \rangle -  \langle b,
      \phi_b^{*}(-g) \rangle - \varepsilon \langle e^{f/\varepsilon},
      e^{-C/\varepsilon} e^{g/\varepsilon}} \rangle

    where :math:`\phi_a(z) = \rho_a z(\log z - 1)` is a scaled entropy, and
    :math:`\phi_a^{*}(z) = \rho_a e^{z/\varepsilon}`, its Legendre transform
    :cite:`sejourne:19`.

    That problem can also be written, instead, using positive scaling vectors
    `u`, `v` of size ``n``, ``m``, handled with the kernel
    :math:`K := e^{-C/\varepsilon}`,

    .. math::

      \arg\max_{u, v >0} - \langle a,\phi_a^{*}(-\varepsilon\log u) \rangle +
      \langle b, \phi_b^{*}(-\varepsilon\log v) \rangle - \langle u, K v \rangle

    Both of these problems corresponds, in their *primal* formulation, to
    solving the unbalanced optimal transport problem with a variable matrix
    :math:`P` of size ``n`` x ``m``:

    .. math::

      \arg\min_{P>0} \langle P,C \rangle +\varepsilon \text{KL}(P | ab^T)
      + \rho_a \text{KL}(P\mathbf{1}_m | a) + \rho_b \text{KL}(P^T \mathbf{1}_n
      | b)

    where :math:`KL` is the generalized Kullback-Leibler divergence.

    The very same primal problem can also be written using a kernel :math:`K`
    instead of a cost :math:`C` as well:

    .. math::

      \arg\min_{P} \varepsilon \text{KL}(P|K)
      + \rho_a \text{KL}(P\mathbf{1}_m | a) +
      \rho_b \text{KL}(P^T \mathbf{1}_n | b)

    The *original* OT problem taught in linear programming courses is recovered
    by using the formulation above relying on the cost :math:`C`, and letting
    :math:`\varepsilon \rightarrow 0`, and :math:`\rho_a, \rho_b \rightarrow
    \infty`.
    In that case the entropy disappears, whereas the :math:`KL` regularization
    above become constraints on the marginals of :math:`P`: This results in a
    standard min cost flow problem. This problem is not handled for now in this
    toolbox, which focuses exclusively on the case :math:`\varepsilon > 0`.

    The *balanced* regularized OT problem is recovered for finite
    :math:`\varepsilon > 0` but letting :math:`\rho_a, \rho_b \rightarrow
    \infty`. This problem can be shown to be equivalent to a matrix scaling
    problem, which can be solved using the Sinkhorn fixed-point algorithm.
    To handle the case :math:`\rho_a, \rho_b \rightarrow \infty`, the
    ``sinkhorn`` function uses parameters ``tau_a`` and ``tau_b`` equal
    respectively to :math:`\rho_a /(\varepsilon + \rho_a)` and
    :math:`\rho_b / (\varepsilon + \rho_b)` instead. Setting either of these
    parameters to 1 corresponds to setting the corresponding
    :math:`\rho_a, \rho_b` to :math:`\infty`.

    The Sinkhorn algorithm solves the reg-OT problem by seeking optimal
    :math:`f`, :math:`g` potentials (or alternatively their parameterization
    as positive scaling vectors :math:`u`, :math:`v`), rather than solving the
    primal problem in :math:`P`. This is mostly for efficiency (potentials and
    scalings have a ``n + m`` memory footprint, rather than ``n m`` required
    to store `P`). This is also because both problems are, in fact, equivalent,
    since the optimal transport :math:`P^{\star}` can be recovered from
    optimal potentials :math:`f^{\star}`, :math:`g^{\star}` or scaling
    :math:`u^{\star}`, :math:`v^{\star}`, using the geometry's cost or kernel
    matrix respectively:

    .. math::

      P^{\star} = \exp\left(\frac{f^{\star}\mathbf{1}_m^T + \mathbf{1}_n g^{*T}-
      C}{\varepsilon}\right) \text{ or } P^{\star} = \text{diag}(u^{\star}) K
      \text{diag}(v^{\star})

    By default, the Sinkhorn algorithm solves this dual problem in :math:`f, g`
    or :math:`u, v` using block coordinate ascent, i.e. devising an update for
    each :math:`f` and :math:`g` (resp. :math:`u` and :math:`v`) that cancels
    their respective gradients, one at a time. These two iterations are repeated
    ``inner_iterations`` times, after which the norm of these gradients will be
    evaluated and compared with the ``threshold`` value. The iterations are then
    repeated as long as that error exceeds ``threshold``.

  Note on Sinkhorn updates:
    The boolean flag ``lse_mode`` sets whether the algorithm is run in either:

    - log-sum-exp mode (``lse_mode=True``), in which case it is directly
      defined in terms of updates to `f` and `g`, using log-sum-exp
      computations. This requires access to the cost matrix :math:`C`, as it is
      stored, or possibly computed on the fly by ``geom``.

    - kernel mode (``lse_mode=False``), in which case it will require access
      to a matrix vector multiplication operator :math:`z \rightarrow K z`,
      where :math:`K` is either instantiated from :math:`C` as
      :math:`\exp(-C/\varepsilon)`, or provided directly. In that case, rather
      than optimizing on :math:`f` and :math:`g`, it is more convenient to
      optimize on their so called scaling formulations,
      :math:`u := \exp(f / \varepsilon)` and :math:`v := \exp(g / \varepsilon)`.
      While faster (applying matrices is faster than applying ``lse`` repeatedly
      over lines), this mode is also less stable numerically, notably for
      smaller :math:`\varepsilon`.

    In the source code, the variables ``f_u`` or ``g_v`` can be either regarded
    as potentials (real) or scalings (positive) vectors, depending on the choice
    of ``lse_mode`` by the user. Once optimization is carried out, we only
    return dual variables in potential form, i.e. ``f`` and ``g``.

    In addition to standard Sinkhorn updates, the user can also use heavy-ball
    type updates using a ``momentum`` parameter in ]0,2[. We also implement a
    strategy that tries to set that parameter adaptively at
    ``chg_momentum_from`` iterations, as a function of progress in the error,
    as proposed in the literature.

    Another upgrade to the standard Sinkhorn updates provided to the users lies
    in using Anderson acceleration. This can be parameterized by setting the
    otherwise null ``anderson`` to a positive integer. When selected,the
    algorithm will recompute, every ``refresh_anderson_frequency`` (set by
    default to 1) an extrapolation of the most recently computed ``anderson``
    iterates. When using that option, notice that differentiation (if required)
    can only be carried out using implicit differentiation, and that all
    momentum related parameters are ignored.

    The ``parallel_dual_updates`` flag is set to ``False`` by default. In that
    setting, ``g_v`` is first updated using the latest values for ``f_u`` and
    ``g_v``, before proceeding to update ``f_u`` using that new value for
    ``g_v``. When the flag is set to ``True``, both ``f_u`` and ``g_v`` are
    updated simultaneously. Note that setting that choice to ``True`` requires
    using some form of averaging (e.g. ``momentum=0.5``). Without this, and on
    its own ``parallel_dual_updates`` won't work.

  Differentiation:
    The optimal solutions ``f`` and ``g`` and the optimal objective
    (``reg_ot_cost``) outputted by the Sinkhorn algorithm can be differentiated
    w.r.t. relevant inputs ``geom``, ``a`` and ``b``. In the default setting,
    implicit differentiation of the optimality conditions (``implicit_diff``
    not equal to ``None``), this has two consequences, treating ``f`` and ``g``
    differently from ``reg_ot_cost``.

    - The termination criterion used to stop Sinkhorn (cancellation of
      gradient of objective w.r.t. ``f_u`` and ``g_v``) is used to differentiate
      ``f`` and ``g``, given a change in the inputs. These changes are computed
      by solving a linear system. The arguments starting with
      ``implicit_solver_*`` allow to define the linear solver that is used, and
      to control for two types or regularization (we have observed that,
      depending on the architecture, linear solves may require higher ridge
      parameters to remain stable). The optimality conditions in Sinkhorn can be
      analyzed as satisfying a ``z=z'`` condition, which are then
      differentiated. It might be beneficial (e.g., as in :cite:`cuturi:20a`)
      to use a preconditioning function ``precondition_fun`` to differentiate
      instead ``h(z) = h(z')``.

    - The objective ``reg_ot_cost`` returned by Sinkhorn uses the so-called
      envelope (or Danskin's) theorem. In that case, because it is assumed that
      the gradients of the dual variables ``f_u`` and ``g_v`` w.r.t. dual
      objective are zero (reflecting the fact that they are optimal), small
      variations in ``f_u`` and ``g_v`` due to changes in inputs (such as
      ``geom``, ``a`` and ``b``) are considered negligible. As a result,
      ``stop_gradient`` is applied on dual variables ``f_u`` and ``g_v`` when
      evaluating the ``reg_ot_cost`` objective. Note that this approach is
      `invalid` when computing higher order derivatives. In that case the
      ``use_danskin`` flag must be set to ``False``.

    An alternative yet more costly way to differentiate the outputs of the
    Sinkhorn iterations is to use unrolling, i.e. reverse mode differentiation
    of the Sinkhorn loop. This is possible because Sinkhorn iterations are
    wrapped in a custom fixed point iteration loop, defined in
    ``fixed_point_loop``, rather than a standard while loop. This is to ensure
    the end result of this fixed point loop can also be differentiated, if
    needed, using standard JAX operations. To ensure differentiability,
    the ``fixed_point_loop.fixpoint_iter_backprop`` loop does checkpointing of
    state variables (here ``f_u`` and ``g_v``) every ``inner_iterations``, and
    backpropagates automatically, block by block, through blocks of
    ``inner_iterations`` at a time.

  Note:
    * The Sinkhorn algorithm may not converge within the maximum number of
      iterations for possibly several reasons:

      1. the regularizer (defined as ``epsilon`` in the geometry ``geom``
         object) is too small. Consider either switching to ``lse_mode=True``
         (at the price of a slower execution), increasing ``epsilon``, or,
         alternatively, if you are unable or unwilling to increase  ``epsilon``,
         either increase ``max_iterations`` or ``threshold``.
      2. the probability weights ``a`` and ``b`` do not have the same total
         mass, while using a balanced (``tau_a=tau_b=1.0``) setup.
         Consider either normalizing ``a`` and ``b``, or set either ``tau_a``
         and/or ``tau_b<1.0``.
      3. OOMs issues may arise when storing either cost or kernel matrices that
         are too large in ``geom``. In the case where, the ``geom`` geometry is
         a ``PointCloud``, some of these issues might be solved by setting the
         ``online`` flag to ``True``. This will trigger a re-computation on the
         fly of the cost/kernel matrix.

    * The weight vectors ``a`` and ``b`` can be passed on with coordinates that
      have zero weight. This is then handled by relying on simple arithmetic for
      ``inf`` values that will likely arise (due to :math:`\log 0` when
      ``lse_mode`` is ``True``, or divisions by zero when ``lse_mode`` is
      ``False``). Whenever that arithmetic is likely to produce ``NaN`` values
      (due to ``-inf * 0``, or ``-inf - -inf``) in the forward pass, we use
      ``jnp.where`` conditional statements to carry ``inf`` rather than ``NaN``
      values. In the reverse mode differentiation, the inputs corresponding to
      these 0 weights (a location `x`, or a row in the corresponding cost/kernel
      matrix), and the weight itself will have ``NaN`` gradient values. This is
      reflects that these gradients are undefined, since these points were not
      considered in the optimization and have therefore no impact on the output.

  Args:
    lse_mode: ``True`` for log-sum-exp computations, ``False`` for kernel
      multiplication.
    threshold: tolerance used to stop the Sinkhorn iterations. This is
      typically the deviation between a target marginal and the marginal of the
      current primal solution when either or both tau_a and tau_b are 1.0
      (balanced or semi-balanced problem), or the relative change between two
      successive solutions in the unbalanced case.
    norm_error: power used to define p-norm of error for marginal/target.
    inner_iterations: the Sinkhorn error is not recomputed at each
      iteration but every ``inner_iterations`` instead.
    min_iterations: the minimum number of Sinkhorn iterations carried
      out before the error is computed and monitored.
    max_iterations: the maximum number of Sinkhorn iterations. If
      ``max_iterations`` is equal to ``min_iterations``, Sinkhorn iterations are
      run by default using a :func:`jax.lax.scan` loop rather than a custom,
      unroll-able :func:`jax.lax.while_loop` that monitors convergence.
      In that case the error is not monitored and the ``converged``
      flag will return ``False`` as a consequence.
    momentum: Momentum instance.
    anderson: AndersonAcceleration instance.
    implicit_diff: instance used to solve implicit differentiation. Unrolls
      iterations if None.
    parallel_dual_updates: updates potentials or scalings in parallel if True,
      sequentially (in Gauss-Seidel fashion) if False.
    recenter_potentials: Whether to re-center the dual potentials.
      If the problem is balanced, the ``f`` potential is zero-centered for
      numerical stability. Otherwise, use the approach of :cite:`sejourne:22`
      to achieve faster convergence. Only used when ``lse_mode = True`` and
      ``tau_a < 1`` and ``tau_b < 1``.
    use_danskin: when ``True``, it is assumed the entropy regularized cost
      is evaluated using optimal potentials that are frozen, i.e. whose
      gradients have been stopped. This is useful when carrying out first order
      differentiation, and is only valid (as with ``implicit_differentiation``)
      when the algorithm has converged with a low tolerance.
    initializer: how to compute the initial potentials/scalings. This refers to
      a few possible classes implemented following the template in
      :class:`~ott.initializers.linear.SinkhornInitializer`.
    progress_fn: callback function which gets called during the Sinkhorn
      iterations, so the user can display the error at each iteration,
      e.g., using a progress bar. See :func:`~ott.utils.default_progress_fn`
      for a basic implementation.
    kwargs_init: keyword arguments when creating the initializer.
  """

  def __init__(
      self,
      lse_mode: bool = True,
      threshold: float = 1e-3,
      norm_error: int = 1,
      inner_iterations: int = 10,
      min_iterations: int = 0,
      max_iterations: int = 2000,
      momentum: Optional[acceleration.Momentum] = None,
      anderson: Optional[acceleration.AndersonAcceleration] = None,
      parallel_dual_updates: bool = False,
      recenter_potentials: bool = False,
      use_danskin: Optional[bool] = None,
      implicit_diff: Optional[implicit_lib.ImplicitDiff
                             ] = implicit_lib.ImplicitDiff(),  # noqa: B008
      initializer: Union[Literal["default", "gaussian", "sorting", "subsample"],
                         init_lib.SinkhornInitializer] = "default",
      progress_fn: Optional[ProgressCallbackFn_t] = None,
      kwargs_init: Optional[Mapping[str, Any]] = None,
  ):
    self.lse_mode = lse_mode
    self.threshold = threshold
    self.inner_iterations = inner_iterations
    self.min_iterations = min_iterations
    self.max_iterations = max_iterations
    self._norm_error = norm_error
    self.anderson = anderson
    self.implicit_diff = implicit_diff

    if momentum is not None:
      self.momentum = acceleration.Momentum(
          momentum.start, momentum.error_threshold, momentum.value,
          self.inner_iterations
      )
    else:
      # Use no momentum if using Anderson or unrolling.
      if self.anderson is not None or self.implicit_diff is None:
        self.momentum = acceleration.Momentum(
            inner_iterations=self.inner_iterations
        )
      else:
        # no momentum
        self.momentum = acceleration.Momentum()

    self.parallel_dual_updates = parallel_dual_updates
    self.recenter_potentials = recenter_potentials
    self.initializer = initializer
    self.progress_fn = progress_fn
    self.kwargs_init = {} if kwargs_init is None else kwargs_init

    # Force implicit_differentiation to True when using Anderson acceleration,
    # Reset all momentum parameters to default (i.e. no momentum)
    if anderson:
      self.implicit_diff = (
          implicit_lib.ImplicitDiff()
          if self.implicit_diff is None else self.implicit_diff
      )
      self.momentum = acceleration.Momentum(
          inner_iterations=self.inner_iterations
      )

    # By default, use Danskin theorem to differentiate
    # the objective when using implicit_lib.
    self.use_danskin = ((self.implicit_diff is not None)
                        if use_danskin is None else use_danskin)

  def __call__(
      self,
      ot_prob: linear_problem.LinearProblem,
      init: Tuple[Optional[jnp.ndarray], Optional[jnp.ndarray]] = (None, None),
      rng: Optional[jax.Array] = None,
  ) -> SinkhornOutput:
    """Run Sinkhorn algorithm.

    Args:
      ot_prob: Linear OT problem.
      init: Initial dual potentials/scalings f_u and g_v, respectively.
        Any `None` values will be initialized using the initializer.
      rng: Random number generator key for stochastic initialization.

    Returns:
      The Sinkhorn output.
    """
    rng = utils.default_prng_key(rng)
    initializer = self.create_initializer()
    init_dual_a, init_dual_b = initializer(
        ot_prob, *init, lse_mode=self.lse_mode, rng=rng
    )
    return run(ot_prob, self, (init_dual_a, init_dual_b))

  def lse_step(
      self, ot_prob: linear_problem.LinearProblem, state: SinkhornState,
      iteration: int
  ) -> SinkhornState:
    """Sinkhorn LSE update."""

    def k(tau_i: float, tau_j: float) -> float:
      num = -tau_j * (tau_a - 1) * (tau_b - 1) * (tau_i - 1)
      denom = (tau_j - 1) * (tau_a * (tau_b - 1) + tau_b * (tau_a - 1))
      return num / denom

    def xi(tau_i: float, tau_j: float) -> float:
      k_ij = k(tau_i, tau_j)
      return k_ij / (1.0 - k_ij)

    def smin(
        potential: jnp.ndarray, marginal: jnp.ndarray, tau: float
    ) -> float:
      rho = uf.rho(ot_prob.epsilon, tau)
      return -rho * mu.logsumexp(-potential / rho, b=marginal)

    # only for an unbalanced problems with `tau_{a,b} < 1`
    recenter = (
        self.recenter_potentials and ot_prob.tau_a < 1.0 and ot_prob.tau_b < 1.0
    )
    w = self.momentum.weight(state, iteration)
    tau_a, tau_b = ot_prob.tau_a, ot_prob.tau_b
    old_fu, old_gv = state.fu, state.gv

    if recenter:
      k11, k22 = k(tau_a, tau_a), k(tau_b, tau_b)
      xi12, xi21 = xi(tau_a, tau_b), xi(tau_b, tau_a)

    # update g potential
    new_gv = tau_b * ot_prob.geom.update_potential(
        old_fu, old_gv, jnp.log(ot_prob.b), iteration, axis=0
    )
    if recenter:
      new_gv -= k22 * smin(old_fu, ot_prob.a, tau_a)
      new_gv += xi21 * smin(new_gv, ot_prob.b, tau_b)
    gv = self.momentum(w, old_gv, new_gv, self.lse_mode)

    if not self.parallel_dual_updates:
      old_gv = gv

    # update f potential
    new_fu = tau_a * ot_prob.geom.update_potential(
        old_fu, old_gv, jnp.log(ot_prob.a), iteration, axis=1
    )
    if recenter:
      new_fu -= k11 * smin(old_gv, ot_prob.b, tau_b)
      new_fu += xi12 * smin(new_fu, ot_prob.a, tau_a)
    fu = self.momentum(w, old_fu, new_fu, self.lse_mode)

    return state.set(potentials=(fu, gv))

  def kernel_step(
      self, ot_prob: linear_problem.LinearProblem, state: SinkhornState,
      iteration: int
  ) -> SinkhornState:
    """Sinkhorn multiplicative update."""
    w = self.momentum.weight(state, iteration)
    old_gv = state.gv
    new_gv = ot_prob.geom.update_scaling(
        state.fu, ot_prob.b, iteration, axis=0
    ) ** ot_prob.tau_b
    gv = self.momentum(w, state.gv, new_gv, self.lse_mode)
    new_fu = ot_prob.geom.update_scaling(
        old_gv if self.parallel_dual_updates else gv,
        ot_prob.a,
        iteration,
        axis=1
    ) ** ot_prob.tau_a
    fu = self.momentum(w, state.fu, new_fu, self.lse_mode)
    return state.set(potentials=(fu, gv))

  def one_iteration(
      self, ot_prob: linear_problem.LinearProblem, state: SinkhornState,
      iteration: int, compute_error: bool
  ) -> SinkhornState:
    """Carries out one Sinkhorn iteration.

    Depending on lse_mode, these iterations can be either in:

      - log-space for numerical stability.
      - scaling space, using standard kernel-vector multiply operations.

    Args:
      ot_prob: the transport problem definition
      state: SinkhornState named tuple.
      iteration: the current iteration of the Sinkhorn loop.
      compute_error: flag to indicate this iteration computes/stores an error

    Returns:
      The updated state.
    """
    # When running updates in parallel (Gauss-Seidel mode), old_g_v will be
    # used to update f_u, rather than the latest g_v computed in this loop.
    # Unused otherwise.
    if self.anderson:
      state = self.anderson.update(state, iteration, ot_prob, self.lse_mode)

    if self.lse_mode:  # In lse_mode, run additive updates.
      state = self.lse_step(ot_prob, state, iteration)
    else:
      state = self.kernel_step(ot_prob, state, iteration)

    if self.anderson:
      state = self.anderson.update_history(state, ot_prob, self.lse_mode)

    # re-computes error if compute_error is True, else set it to inf.
    err = jax.lax.cond(
        jnp.logical_or(
            iteration == self.max_iterations - 1,
            jnp.logical_and(compute_error, iteration >= self.min_iterations)
        ),
        lambda state, prob: state.solution_error(
            prob,
            self.norm_error,
            lse_mode=self.lse_mode,
            parallel_dual_updates=self.parallel_dual_updates,
            recenter=self.recenter_potentials
        )[0],
        lambda *_: jnp.array(jnp.inf, dtype=ot_prob.dtype),
        state,
        ot_prob,
    )
    errors = state.errors.at[iteration // self.inner_iterations, :].set(err)
    state = state.set(errors=errors)

    if self.progress_fn is not None:
      jax.debug.callback(
          self.progress_fn,
          (iteration, self.inner_iterations, self.max_iterations, state)
      )
    return state

  def _converged(self, state: SinkhornState, iteration: int) -> bool:
    err = state.errors[iteration // self.inner_iterations - 1, 0]
    return jnp.logical_and(iteration > 0, err < self.threshold)

  def _diverged(self, state: SinkhornState, iteration: int) -> bool:
    err = state.errors[iteration // self.inner_iterations - 1, 0]
    return jnp.logical_not(jnp.isfinite(err))

  def _continue(self, state: SinkhornState, iteration: int) -> bool:
    """Continue while not(converged) and not(diverged)."""
    return jnp.logical_and(
        jnp.logical_not(self._diverged(state, iteration)),
        jnp.logical_not(self._converged(state, iteration))
    )

  @property
  def outer_iterations(self) -> int:
    """Upper bound on number of times inner_iterations are carried out.

    This integer can be used to set constant array sizes to track the algorithm
    progress, notably errors.
    """
    return np.ceil(self.max_iterations / self.inner_iterations).astype(int)

  def init_state(
      self, ot_prob: linear_problem.LinearProblem, init: Tuple[jnp.ndarray,
                                                               jnp.ndarray]
  ) -> SinkhornState:
    """Return the initial state of the loop."""
    errors = -jnp.ones((self.outer_iterations, len(self.norm_error)),
                       dtype=ot_prob.dtype)
    state = SinkhornState(init, errors=errors)
    return self.anderson.init_maps(ot_prob, state) if self.anderson else state

  def output_from_state(
      self, ot_prob: linear_problem.LinearProblem, state: SinkhornState
  ) -> SinkhornOutput:
    """Create an output from a loop state.

    Note:
      When differentiating the regularized OT cost, and assuming Sinkhorn has
      run to convergence, Danskin's (or the envelope)
      `theorem <https://en.wikipedia.org/wiki/Danskin%27s_theorem>`_
      :cite:`danskin:67,bertsekas:71`
      states that the resulting OT cost as a function of the inputs
      (``geometry``, ``a``, ``b``) behaves locally as if the dual optimal
      potentials were frozen and did not vary with those inputs.

      Notice this is only valid, as when using ``implicit_differentiation``
      mode, if the Sinkhorn algorithm outputs potentials that are near optimal.
      namely when the threshold value is set to a small tolerance.

      The flag ``use_danskin`` controls whether that assumption is made. By
      default, that flag is set to the value of ``implicit_differentiation`` if
      not specified. If you wish to compute derivatives of order 2 and above,
      set ``use_danskin`` to ``False``.

    Args:
      ot_prob: the transport problem.
      state: a SinkhornState.

    Returns:
      A SinkhornOutput.
    """
    geom = ot_prob.geom

    f = state.fu if self.lse_mode else geom.potential_from_scaling(state.fu)
    g = state.gv if self.lse_mode else geom.potential_from_scaling(state.gv)
    if self.recenter_potentials:
      f, g = state.recenter(f, g, ot_prob=ot_prob)

    # By convention, the algorithm is said to have converged if the algorithm
    # has not nan'ed during iterations (notice some errors might be infinite,
    # this convention is used when the error is not recomputed), and if the
    # last recorded error is lower than the threshold. Note that this will be
    # the case if either the algorithm terminated earlier (in which case the
    # last state.errors[-1] = -1 by convention) or if the algorithm carried out
    # the maximal number of iterations and its last recorded error (at -1
    # position) is lower than the threshold.

    converged = jnp.logical_and(
        jnp.logical_not(jnp.any(jnp.isnan(state.errors))), state.errors[-1]
        < self.threshold
    )[0]

    return SinkhornOutput((f, g),
                          errors=state.errors[:, 0],
                          threshold=jnp.array(self.threshold),
                          converged=converged,
                          inner_iterations=self.inner_iterations)

  @property
  def norm_error(self) -> Tuple[int, ...]:
    """Powers used to compute the p-norm between marginal/target."""
    # To change momentum adaptively, one needs errors in ||.||_1 norm.
    # In that case, we add this exponent to the list of errors to compute,
    # notably if that was not the error requested by the user.
    if self.momentum and self.momentum.start > 0 and self._norm_error != 1:
      return self._norm_error, 1
    return self._norm_error,

  # TODO(michalk8): in the future, enforce this (+ in GW) via abstract method
  def create_initializer(self) -> init_lib.SinkhornInitializer:  # noqa: D102
    if isinstance(self.initializer, init_lib.SinkhornInitializer):
      return self.initializer
    if self.initializer == "default":
      return init_lib.DefaultInitializer()
    if self.initializer == "gaussian":
      return init_lib.GaussianInitializer()
    if self.initializer == "sorting":
      return init_lib.SortingInitializer(**self.kwargs_init)
    if self.initializer == "subsample":
      return init_lib.SubsampleInitializer(**self.kwargs_init)
    raise NotImplementedError(
        f"Initializer `{self.initializer}` is not yet implemented."
    )

  def tree_flatten(self):  # noqa: D102
    aux = vars(self).copy()
    aux["norm_error"] = aux.pop("_norm_error")
    aux.pop("threshold")
    return [self.threshold], aux

  @classmethod
  def tree_unflatten(cls, aux_data, children):  # noqa: D102
    return cls(**aux_data, threshold=children[0])


def run(
    ot_prob: linear_problem.LinearProblem, solver: Sinkhorn,
    init: Tuple[jnp.ndarray, ...]
) -> SinkhornOutput:
  """Run loop of the solver, outputting a state upgraded to an output."""
  iter_fun = _iterations_implicit if solver.implicit_diff else iterations
  out = iter_fun(ot_prob, solver, init)
  # Be careful here, the geom and the cost are injected at the end, where it
  # does not interfere with the implicit differentiation.
  out = out.set_cost(ot_prob, solver.lse_mode, solver.use_danskin)
  return out.set(ot_prob=ot_prob)


def iterations(
    ot_prob: linear_problem.LinearProblem, solver: Sinkhorn,
    init: Tuple[jnp.ndarray, ...]
) -> SinkhornOutput:
  """Jittable Sinkhorn loop. args contain initialization variables."""

  def cond_fn(
      iteration: int, const: Tuple[linear_problem.LinearProblem, Sinkhorn],
      state: SinkhornState
  ) -> bool:
    _, solver = const
    return solver._continue(state, iteration)

  def body_fn(
      iteration: int, const: Tuple[linear_problem.LinearProblem, Sinkhorn],
      state: SinkhornState, compute_error: bool
  ) -> SinkhornState:
    ot_prob, solver = const
    return solver.one_iteration(ot_prob, state, iteration, compute_error)

  # Run the Sinkhorn loop. Choose either a standard fixpoint_iter loop if
  # differentiation is implicit, otherwise switch to the backprop friendly
  # version of that loop if unrolling to differentiate.
  if solver.implicit_diff:
    fix_point = fixed_point_loop.fixpoint_iter
  else:
    fix_point = fixed_point_loop.fixpoint_iter_backprop

  const = ot_prob, solver
  state = solver.init_state(ot_prob, init)
  state = fix_point(
      cond_fn, body_fn, solver.min_iterations, solver.max_iterations,
      solver.inner_iterations, const, state
  )
  return solver.output_from_state(ot_prob, state)


def _iterations_taped(
    ot_prob: linear_problem.LinearProblem, solver: Sinkhorn,
    init: Tuple[jnp.ndarray, ...]
) -> Tuple[SinkhornOutput, Tuple[jnp.ndarray, jnp.ndarray,
                                 linear_problem.LinearProblem, Sinkhorn]]:
  """Run forward pass of the Sinkhorn algorithm storing side information."""
  state = iterations(ot_prob, solver, init)
  return state, (state.f, state.g, ot_prob, solver)


def _iterations_implicit_bwd(res, gr: SinkhornOutput):
  """Run Sinkhorn in backward mode, using implicit differentiation.

  Args:
    res: residual data sent from fwd pass, used for computations below. In this
      case consists in the output itself, as well as inputs against which we
      wish to differentiate.
    gr: gradients w.r.t outputs of fwd pass, here w.r.t size f, g, errors. Note
      that differentiability w.r.t. errors is not handled, and only f, g is
      considered.

  Returns:
    a tuple of gradients: PyTree for geom, one jnp.ndarray for each of a and b.
  """
  f, g, ot_prob, solver = res
  out = solver.implicit_diff.gradient(
      ot_prob, f, g, solver.lse_mode, gr.potentials
  )
  return *out, None, None


# sets threshold, norm_errors, geom, a and b to be differentiable, as those are
# non-static. Only differentiability w.r.t. geom, a and b will be used.
_iterations_implicit = jax.custom_vjp(iterations)
_iterations_implicit.defvjp(_iterations_taped, _iterations_implicit_bwd)
